import json
import os
import concurrent.futures
from typing import List, Dict, Any
import byzerllm
from byzerllm.apps.byzer_storage.generate_sft_data import to_qa_pairs, read_alpaca_zh


def train_sft(
    name: str,
    memories: List[str],
    options: Dict[str, Any],
    dataset_dir: str,
    loras_dir: str,
    target_model_path: str,
    data_model_name: str,
    model_name: str,
):
    data = []
    min_samples = options.pop("min_samples", 1000)

    if memories:
        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
            futures = []
            llm = byzerllm.ByzerLLM().from_default_model(data_model_name)
            for memory in memories:
                futures.append(executor.submit(to_qa_pairs, memory, llm))

            for future in concurrent.futures.as_completed(futures):
                qa_pairs = future.result()
                for qa_pair in qa_pairs:
                    item = {
                        "instruction": qa_pair.question,
                        "input": "",
                        "output": qa_pair.answer,
                    }
                    data.append(item)
    else:
        v = read_alpaca_zh()
        data.extend(v)

    if len(data) < min_samples:
        data = data * (min_samples // len(data) + 1)

    with open(f"{dataset_dir}/data.json", "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2, ensure_ascii=False)

    with open(f"{dataset_dir}/dataset_info.json", "w", encoding="utf-8") as f:
        r = {"data": {"file_name": "data.json"}}
        f.write(json.dumps(r, indent=2, ensure_ascii=False))

    args = dict(
        stage="sft",
        do_train=True,
        model_name_or_path=target_model_path,
        dataset="data",
        dataset_dir=dataset_dir,
        cutoff_len=1024,
        max_samples=1000000,
        overwrite_cache=True,
        preprocessing_num_workers=1,
        template="llama3",
        finetuning_type="lora",
        lora_target="all",
        output_dir=f"{loras_dir}/{name}",
        overwrite_output_dir=True,
        per_device_train_batch_size=2,
        gradient_accumulation_steps=4,
        lr_scheduler_type="cosine",
        logging_steps=10,
        warmup_ratio=0.1,
        save_steps=1000,
        plot_loss=True,
        learning_rate=5e-5,
        num_train_epochs=1000.0,
        max_grad_norm=1.0,
        quantization_bit=4,
        loraplus_lr_ratio=16.0,
        fp16=True,
        ddp_timeout=180000000,
    )
    os.environ["WANDB_DISABLED"] = "true"
    from llamafactory.train import tuner

    tuner.run_exp({**args, **options})




"""
Tag Parser Module - Robust regex-based tag parsing with error handling
"""

import re
from typing import List, Dict, Any, Optional, Tuple, Callable
from dataclasses import dataclass, field
import logging

logger = logging.getLogger(__name__)


class ParseError(Exception):
    """Custom exception for parsing errors"""
    def __init__(self, message: str, position: Optional[int] = None):
        self.message = message
        self.position = position
        super().__init__(self.message)


@dataclass
class Tag:
    """Represents a parsed tag with its content"""
    name: str
    content: str
    start_pos: int
    end_pos: int
    attributes: Dict[str, str] = field(default_factory=dict)


class TagParser:
    """
    Robust regex-based tag parser that extracts tagged content from text.
    
    Supports multiple tag formats:
    - <_tag_>content</_tag_>
    - <tag>content</tag>
    - Custom tag patterns
    """
    
    def __init__(self, 
                 tag_pattern: Optional[str] = None,
                 strict_mode: bool = True,
                 max_tag_length: int = 10_000_000):  # 10MB limit for base64 data
        """
        Initialize the parser with optional custom tag pattern.
        
        Args:
            tag_pattern: Custom regex pattern for tags (default supports <_tag_> and <tag>)
            strict_mode: If True, raises exceptions on malformed tags
            max_tag_length: Maximum allowed tag content length (default: 10MB for base64 data)
        """
        self.strict_mode = strict_mode
        self.max_tag_length = max_tag_length
        
        # Default pattern supports both <_tag_> and <tag> formats
        if tag_pattern is None:
            self.tag_pattern = r'<(_?\w+_?)>(.*?)</\1>'
        else:
            self.tag_pattern = tag_pattern
            
        try:
            self.compiled_pattern = re.compile(self.tag_pattern, re.DOTALL)
        except re.error as e:
            raise ParseError(f"Invalid regex pattern: {e}")
    
    def parse(self, text: str) -> List[Tag]:
        """
        Parse all tags from the input text.
        
        Args:
            text: Input text containing tags
            
        Returns:
            List of Tag objects
            
        Raises:
            ParseError: If strict_mode is True and malformed tags are found
        """
        if not text:
            return []
            
        tags = []
        
        try:
            for match in self.compiled_pattern.finditer(text):
                tag_name = match.group(1)
                content = match.group(2)
                
                # Validate tag content length
                if len(content) > self.max_tag_length:
                    error_msg = f"Tag content exceeds maximum length of {self.max_tag_length}"
                    if self.strict_mode:
                        raise ParseError(error_msg, match.start())
                    else:
                        logger.warning(f"{error_msg} at position {match.start()}")
                        continue
                
                tag = Tag(
                    name=tag_name,
                    content=content.strip(),
                    start_pos=match.start(),
                    end_pos=match.end()
                )
                tags.append(tag)
                
        except re.error as e:
            raise ParseError(f"Invalid regex pattern: {str(e)}")
            
        return tags
    
    def extract_by_tag_name(self, text: str, tag_name: str) -> List[str]:
        """
        Extract all content for a specific tag name.
        
        Args:
            text: Input text
            tag_name: Name of the tag to extract
            
        Returns:
            List of content strings for the specified tag
        """
        tags = self.parse(text)
        return [tag.content for tag in tags if tag.name == tag_name]
    
    def replace_tags(self, 
                     text: str, 
                     replacer_func: Callable[[Tag], str],
                     tag_filter: Optional[Callable[[Tag], bool]] = None) -> str:
        """
        Replace tags in text using a custom replacer function.
        
        Args:
            text: Input text
            replacer_func: Function that takes a Tag and returns replacement string
            tag_filter: Optional function to filter which tags to replace
            
        Returns:
            Text with tags replaced
        """
        tags = self.parse(text)
        
        # Sort tags by position in reverse to avoid offset issues
        tags.sort(key=lambda t: t.start_pos, reverse=True)
        
        result = text
        for tag in tags:
            if tag_filter is None or tag_filter(tag):
                replacement = replacer_func(tag)
                result = result[:tag.start_pos] + replacement + result[tag.end_pos:]
                
        return result
    
    def validate_tags(self, text: str) -> Tuple[bool, List[str]]:
        """
        Validate that all tags in the text are properly formed.
        
        Args:
            text: Input text to validate
            
        Returns:
            Tuple of (is_valid, list_of_errors)
        """
        errors = []
        
        # Check for unclosed tags
        opening_pattern = r'<(_?\w+_?)>'
        closing_pattern = r'</(_?\w+_?)>'
        
        opening_tags = re.findall(opening_pattern, text)
        closing_tags = re.findall(closing_pattern, text)
        
        # Count occurrences
        opening_counts = {}
        closing_counts = {}
        
        for tag in opening_tags:
            opening_counts[tag] = opening_counts.get(tag, 0) + 1
            
        for tag in closing_tags:
            closing_counts[tag] = closing_counts.get(tag, 0) + 1
            
        # Check for mismatches
        all_tags = set(opening_counts.keys()) | set(closing_counts.keys())
        
        for tag in all_tags:
            open_count = opening_counts.get(tag, 0)
            close_count = closing_counts.get(tag, 0)
            
            if open_count != close_count:
                errors.append(f"Tag '{tag}' has {open_count} opening tags but {close_count} closing tags")
                
        # Check for nested tags of the same type (which would break the regex)
        for tag_name in opening_counts:
            pattern = f'<{tag_name}>.*?<{tag_name}>.*?</{tag_name}>.*?</{tag_name}>'
            if re.search(pattern, text, re.DOTALL):
                errors.append(f"Nested tags of type '{tag_name}' detected")
                
        return len(errors) == 0, errors
    
    def extract_with_positions(self, text: str) -> List[Dict[str, Any]]:
        """
        Extract tags with their positions in the original text.
        
        Returns:
            List of dictionaries containing tag info and positions
        """
        tags = self.parse(text)
        return [
            {
                'name': tag.name,
                'content': tag.content,
                'start': tag.start_pos,
                'end': tag.end_pos,
                'full_tag': text[tag.start_pos:tag.end_pos]
            }
            for tag in tags
        ]




from byzerllm.utils.client import ByzerLLM,InferBackend,Templates
import ray
import pytest

model_path = "/home/winubuntu/models/deepseek-coder-1.3b-instruct"
model_name = "chat"

class TestByzerLLMVLLMDeploy(object):
    llm = None

    def setup_class(self):
        if ray.is_initialized():
            ray.shutdown()
        
        ray.init(address="auto",namespace="default")
        
        self.llm = ByzerLLM(verbose=True)

        if self.llm.is_model_exist(model_name):
            self.llm.undeploy(model_name)

        self.llm.setup_gpus_per_worker(1).setup_num_workers(1).setup_infer_backend(InferBackend.VLLM)
        self.llm.deploy(
            model_path=model_path,
            pretrained_model_type="custom/auto",
            udf_name=model_name,
            infer_params={}
        )
        self.llm.setup_default_model_name(model_name)        
        
        
    def teardown_class(self):
        if self.llm.is_model_exist(model_name):
            self.llm.undeploy(model_name)

    def test_template(self):
        import json
        meta = self.llm.get_meta(model=model_name)        
        assert meta.get("support_chat_template",False) == True
        m = self.llm.apply_chat_template(model_name,json.dumps([{
            "content":"You are an AI programming assistant, utilizing the DeepSeek Coder model, developed by DeepSeek Company, and you only answer questions related to computer science. For politically sensitive questions, security and privacy issues, and other non-computer science questions, you will refuse to answer.",
            "role":"system"        
        },
            {
            "content":"写个冒泡算法",
            "role":"user"
        }],ensure_ascii=False))
        print(m,flush=True)
    
    def test_chat_oai(self):
        self.llm.setup_template(model_name,Templates.deepseek_code_chat())
        t = self.llm.chat_oai([{
            "content":"You are an AI programming assistant, utilizing the DeepSeek Coder model, developed by DeepSeek Company, and you only answer questions related to computer science. For politically sensitive questions, security and privacy issues, and other non-computer science questions, you will refuse to answer.",
            "role":"system"        
        },
            {
            "content":"写个冒泡算法",
            "role":"user"
        }])
        print("======",t)

    def test_insert(self):
        self.llm.setup_template(model_name,Templates.deepseek_code_insertion())
        t = self.llm.chat_oai([{
            "content":'''
def quick_sort(arr):
    if len(arr) <= 1:
        return arr
    pivot = arr[0]
    left = []
    right = []
<｜fim▁hole｜>
        if arr[i] < pivot:
            left.append(arr[i])
        else:
            right.append(arr[i])
    return quick_sort(left) + [pivot] + quick_sort(right)
''',
            "role":"system"        
        }])
        print("======",t)  

    def test_completion(self):
        self.llm.setup_template(model_name,Templates.deepseek_code_completion())
        t = self.llm.chat_oai([{
            "content":'''
#write a quick sort algorithm
''',
            "role":"system"        
        }])
        print("======",t)      
    



    



from http import HTTPStatus
from typing import List, Dict
import uuid
import google.generativeai as genai
from google.generativeai.types import content_types
import time
import ray
from byzerllm.utils import BlockVLLMStreamServer,StreamOutputs,SingleOutput,SingleOutputMeta
import threading
import asyncio
from byzerllm.utils.langutil import asyncfy_with_semaphore


class CustomSaasAPI:
    def __init__(self, infer_params: Dict[str, str]) -> None:
        self.api_key: str = infer_params["saas.api_key"]  
        self.model = infer_params.get("saas.model", "gemini-pro")
        self.meta = {
            "model_deploy_type": "saas",
            "backend":"saas",
            "support_stream": True
        } 
        genai.configure(api_key=self.api_key)
        self.client = genai.GenerativeModel(self.model)
                
        try:
            ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        except ValueError:            
            ray.remote(BlockVLLMStreamServer).options(name="BLOCK_VLLM_STREAM_SERVER",lifetime="detached",max_concurrency=1000).remote()     

     # saas/proprietary
    def get_meta(self):
        return [self.meta] 
    
    async def async_get_meta(self):
        return await asyncfy_with_semaphore(self.get_meta)()

    async def async_stream_chat(
            self,
            tokenizer,
            ins: str,
            his: List[dict] = [],
            max_length: int = 1024,
            top_p: float = 0.9,
            temperature: float = 0.1,
            **kwargs
    ):
                
        start_time = time.monotonic()

        other_params = {}                
        if "stream" in kwargs:        
            other_params["stream"] = kwargs["stream"]

        stream = kwargs.get("stream",False)    
        
        new_messages = []

        for message in his:
            role = message["role"]
            if role == "assistant":
                role = "model"
            new_messages.append({"role":role,"content":content_types.to_content(message["content"])})
        
        new_messages.append({"role":"user","content":content_types.to_content(ins)})   
        
        res_data = await asyncfy_with_semaphore(lambda:self.client.generate_content(contents=new_messages,stream=stream))()
        
        if stream:            
            server = ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
            request_id = [None]
           
            def writer(): 
                r = ""
                for response in res_data:                                        
                    v = response.text
                    r += v
                    request_id[0] = str(uuid.uuid4())                        
                    ray.get(server.add_item.remote(request_id[0], 
                                                    StreamOutputs(outputs=[SingleOutput(text=r,metadata=SingleOutputMeta(
                                                        input_tokens_count=0,
                                                        generated_tokens_count=0,
                                                    ))])
                                                    ))
                    
                ray.get(server.mark_done.remote(request_id[0]))

            threading.Thread(target=writer,daemon=True).start()            
                               
            time_count= 10*100
            while request_id[0] is None and time_count > 0:
                await asyncio.sleep(0.01)
                time_count -= 1
            
            if request_id[0] is None:
                raise Exception("Failed to get request id")
            
            def write_running():
                return ray.get(server.add_item.remote(request_id[0], "RUNNING"))
                        
            await asyncio.to_thread(write_running)
            return [("",{"metadata":{"request_id":request_id[0],"stream_server":"BLOCK_VLLM_STREAM_SERVER"}})]  
              
        time_cost = time.monotonic() - start_time
        
        
        generated_text = res_data.text
        generated_tokens_count = 0
        input_tokens_count = 0

        return [(generated_text,{"metadata":{
                "request_id":res_data["request_id"],
                "input_tokens_count":input_tokens_count,
                "generated_tokens_count":generated_tokens_count,
                "time_cost":time_cost,
                "first_token_time":0,
                "speed":float(generated_tokens_count)/time_cost,        
            }})] 

    
        




import byzerllm
from typing import Union, List, Optional, Dict

from .schema import LLMModel
from .registry import ModelRegistry
from .factory import LLMFactory
from .pricing import PricingManager


class LLMManager:
    """大语言模型管理器，统一管理所有 LLM 的配置、实例化和操作"""
    
    def __init__(self, models_json_path: Optional[str] = None):
        """
        初始化 LLM 管理器
        
        Args:
            models_json_path: 模型配置文件路径，默认为 ~/.auto-coder/keys/models.json
        """
        self.registry = ModelRegistry(models_json_path)
        self.pricing = PricingManager(self.registry)
        # 立即加载模型以初始化缓存
        self.registry.load()
    
    def get_model(self, name: str) -> Optional[LLMModel]:
        """
        获取模型配置信息
        
        Args:
            name: 模型名称
            
        Returns:
            LLMModel 对象，如果不存在则返回 None
        """
        return self.registry.get(name)
    
    def get_all_models(self) -> Dict[str, LLMModel]:
        """获取所有模型配置"""
        return self.registry.get_all()
    
    def get_model_info(self, name: str, product_mode: str) -> Optional[Dict]:
        """
        获取模型信息（兼容旧版本接口）
        
        Args:
            name: 模型名称
            product_mode: 产品模式（pro 或 lite）
            
        Returns:
            模型信息字典，如果不存在则返回 None
        """
        if product_mode == "pro":
            # Pro 模式下直接返回 None（与原实现保持一致）
            return None
        
        model = self.get_model(name)
        if not model:
            return None
        
        # 转换为字典格式，保持兼容性
        return {
            "name": model.name,
            "description": model.description,
            "model_name": model.model_name,
            "model_type": model.model_type,
            "base_url": model.base_url,
            "api_key": model.get_api_key(),
            "is_reasoning": model.is_reasoning,
            "input_price": model.input_price,
            "output_price": model.output_price,
            "max_output_tokens": model.max_output_tokens,
            "context_window": model.context_window
        }
    
    def get_single_llm(self, model_names: str, product_mode: str) -> Optional[Union[byzerllm.ByzerLLM, byzerllm.SimpleByzerLLM]]:
        """
        根据模型名称和产品模式获取 LLM 实例
        
        Args:
            model_names: 逗号分隔的模型名称列表
            product_mode: 产品模式（pro 或 lite）
            
        Returns:
            配置好的 LLM 实例
            
        Raises:
            ValueError: 当所有指定的模型都无法创建时，包含详细的错误信息
        """
        models = list(self.registry.get_all().values())
        return LLMFactory.create_llm_from_names(model_names, models, product_mode)
    
    def get_llm_names(self, llm: Union[byzerllm.ByzerLLM, byzerllm.SimpleByzerLLM, str, List],
                      target_model_type: Optional[str] = None) -> List[str]:
        """
        获取 LLM 的模型名称列表
        
        Args:
            llm: LLM 实例或模型名称
            target_model_type: 目标模型类型（可选）
            
        Returns:
            模型名称列表
        """
        return LLMFactory.get_llm_names(llm, target_model_type)
    
    def add_models(self, models: List[Union[LLMModel, Dict]]) -> None:
        """
        批量添加模型
        
        Args:
            models: 模型列表，可以是 LLMModel 对象或字典
        """
        for model_data in models:
            if isinstance(model_data, dict):
                # 如果是字典，转换为 LLMModel
                model = LLMModel(**model_data)
            else:
                model = model_data
            
            # 保存 api_key 的副本（如果有的话）
            api_key_to_save = model.api_key
            
            # 如果有 api_key，设置 api_key_path（如果没有的话）
            if api_key_to_save and not model.api_key_path:
                model.api_key_path = model.name.replace("/", "_")
            
            # 清除内存中的 api_key，以便正确保存到文件
            model.api_key = None
            
            # 首先添加模型配置
            self.registry.add_or_update(model)
            
            # 如果有 api_key，再保存密钥
            if api_key_to_save and model.api_key_path:
                # 使用文件锁保存密钥到文件
                api_key_file = self.registry.models_json_path.parent / model.api_key_path
                self.registry._ensure_dir()
                
                # 导入文件锁上下文管理器
                from filelock import FileLock
                
                with FileLock(self.registry._get_lock_path(api_key_file), timeout=5):
                    with open(api_key_file, 'w', encoding='utf-8') as f:
                        f.write(api_key_to_save.strip())
    
    def update_model_with_api_key(self, name: str, api_key: str) -> bool:
        """
        更新模型的 API 密钥
        
        Args:
            name: 模型名称
            api_key: API 密钥
            
        Returns:
            是否成功更新
        """
        return self.registry.save_api_key(name, api_key)
    
    def update_model(self, name: str, model_data: Dict) -> Optional[LLMModel]:
        """
        更新模型信息
        
        Args:
            name: 要更新的模型名称
            model_data: 包含新信息的字典
            
        Returns:
            更新后的模型，如果未找到则返回 None
        """
        model = self.registry.get(name)
        if not model:
            return None
        
        # 更新字段
        if "description" in model_data:
            model.description = model_data["description"]
        if "model_name" in model_data:
            model.model_name = model_data["model_name"]
        if "model_type" in model_data:
            model.model_type = model_data["model_type"]
        if "base_url" in model_data:
            model.base_url = model_data["base_url"]
        if "is_reasoning" in model_data:
            model.is_reasoning = model_data["is_reasoning"]
        if "input_price" in model_data:
            model.input_price = float(model_data["input_price"])
        if "output_price" in model_data:
            model.output_price = float(model_data["output_price"])
        if "max_output_tokens" in model_data:
            model.max_output_tokens = int(model_data["max_output_tokens"])
        if "context_window" in model_data:
            model.context_window = int(model_data["context_window"])
        
        # 保存更新
        self.registry.add_or_update(model)
        
        # 如果提供了 API 密钥，单独处理
        if "api_key" in model_data and model_data["api_key"]:
            self.update_model_with_api_key(name, model_data["api_key"])
        
        return model
    
    def has_key(self, name: str) -> bool:
        """
        检查模型是否已配置密钥
        
        Args:
            name: 模型名称
            
        Returns:
            是否已配置密钥
        """
        model = self.get_model(name)
        return model.has_api_key if model else False
    
    def check_model_exists(self, name: str) -> bool:
        """
        检查模型是否存在
        
        Args:
            name: 模型名称
            
        Returns:
            是否存在
        """
        return self.get_model(name) is not None
    
    # 价格管理相关方法
    def update_input_price(self, name: str, price: float) -> bool:
        """更新模型输入价格"""
        return self.pricing.update_input_price(name, price)
    
    def update_output_price(self, name: str, price: float) -> bool:
        """更新模型输出价格"""
        return self.pricing.update_output_price(name, price)
    
    def update_price(self, name: str, input_price: Optional[float] = None,
                    output_price: Optional[float] = None) -> bool:
        """同时更新输入和输出价格"""
        return self.pricing.update_prices(name, input_price, output_price)
    
    def get_cost_estimate(self, name: str, input_tokens: int, output_tokens: int) -> Optional[float]:
        """估算使用成本"""
        return self.pricing.get_cost_estimate(name, input_tokens, output_tokens)
    
    def remove_model(self, name: str) -> bool:
        """
        删除模型
        
        Args:
            name: 模型名称
            
        Returns:
            是否成功删除
        """
        return self.registry.remove_model(name) 


import ray
from ray.types import ObjectRef
from byzerllm.records import (
    ClusterSettings,
    EnvSettings,
    JVMSettings,
    TableSettings,
    SearchQuery,
    ResourceRequirementSettings,
    ResourceRequirement,
)
from typing import List, Dict, Any, Optional, Union
import byzerllm.utils.object_store_ref_util as ref_utils
import json
from loguru import logger


class ClusterBuilder:

    def __init__(self, br: "ByzerRetrievalProxy",pure_client:bool=False) -> None:
        self.name = None
        self.location = None
        self.numNodes = 1
        self.nodeMemory = "2g"
        self.nodeCPU = 1
        self.enableZGC = True
        self.javaHome = None
        self.path = None

        self.cluster_settings = None
        self.env_settings = None
        self.jvm_settings = None
        self.resource_requirement_settings = None

        self.custom_resources = {}

        self.br = br
        self.pure_client = pure_client

    def set_name(self, name: str):
        self.name = name
        return self

    def set_location(self, location: str):
        self.location = location
        return self

    def set_num_nodes(self, numNodes: int):
        self.numNodes = numNodes
        return self

    def set_node_memory(self, nodeMemory: str):
        self.nodeMemory = nodeMemory
        return self

    def set_custom_resource(self, k: str, v: float):
        self.custom_resources[k] = v
        return self

    def set_node_cpu(self, nodeCPU: int):
        self.nodeCPU = nodeCPU
        return self

    def set_enable_zgc(self):
        self.enableZGC = True
        return self

    def set_java_home(self, javaHome: str):
        self.javaHome = javaHome
        return self

    def set_path(self, path: str):
        self.path = path
        return self

    def build(self):

        if self.name is None:
            raise Exception("name is required")

        if self.location is None:
            raise Exception("location is required")

        self.cluster_settings = ClusterSettings(self.name, self.location, self.numNodes)

        if self.javaHome is None:
            raise Exception("javaHome is required")

        if self.path is None:
            raise Exception("path is required")

        self.env_settings = EnvSettings(javaHome=self.javaHome, path=self.path)

        jvmOptions = []
        resourceOptions = []
        if self.enableZGC:
            jvmOptions.append("-XX:+UseZGC")

        if self.nodeMemory:
            jvmOptions.append(f"-Xmx{self.nodeMemory}")

        if self.nodeCPU:
            resourceOptions.append(ResourceRequirement("CPU", self.nodeCPU))

        if self.custom_resources:
            for k, v in self.custom_resources.items():
                resourceOptions.append(ResourceRequirement(k, v))

        self.jvm_settings = JVMSettings(jvmOptions)
        self.resource_requirement_settings = ResourceRequirementSettings(
            resourceOptions
        )

    def start_cluster(self) -> bool:
        self.build() 
        if self.pure_client:       
            return ray.get(
                self.br.start_cluster.remote(
                    self.cluster_settings,
                    self.env_settings,
                    self.jvm_settings,
                    self.resource_requirement_settings,
                )
            )
        
        return self.br.start_cluster(
            self.cluster_settings,
            self.env_settings,
            self.jvm_settings,
            self.resource_requirement_settings,
        )


class ByzerRetrievalProxy:

    def __init__(self):
        self.launched = False
        self.retrieval_gateway = None
        self.clusters = {}

    def launch_gateway(self) -> ray.actor.ActorHandle:

        try:
            self.retrieval_gateway = ray.get_actor("RetrievalGateway")
        except Exception:
            pass

        if self.retrieval_gateway:
            self.launched = True
            return self.retrieval_gateway

        if self.launched:
            return ray.get_actor("RetrievalGateway")

        retrieval_gateway_launcher_clzz = ray.cross_language.java_actor_class(
            "tech.mlsql.retrieval.RetrievalGatewayLauncher"
        )
        retrieval_gateway_launcher = retrieval_gateway_launcher_clzz.remote()
        ray.get(retrieval_gateway_launcher.launch.remote())
        retrieval_gateway = ray.get_actor("RetrievalGateway")
        self.launched = True
        self.retrieval_gateway = retrieval_gateway
        return retrieval_gateway

    def gateway(slef) -> ray.actor.ActorHandle:
        return ray.get_actor("RetrievalGateway")

    def cluster_builder(self) -> ClusterBuilder:
        br = self
        return ClusterBuilder(br,pure_client=False)

    def start_cluster(
        self,
        cluster_settings: ClusterSettings,
        env_settings: EnvSettings,
        jvm_settings: JVMSettings,
        resource_requirement_settings: ResourceRequirementSettings = ResourceRequirementSettings(
            []
        ),
    ) -> bool:
        if not self.launched:
            raise Exception("Please launch gateway first")

        if cluster_settings.name in self.clusters:
            raise Exception(f"Cluster {cluster_settings.name} already exists")

        try:
            ray.get_actor(cluster_settings.name)
            raise Exception(f"Cluster {cluster_settings.name} already exists")
        except ValueError:
            pass

        obj_ref1 = self.retrieval_gateway.buildCluster.remote(
            cluster_settings.json(),
            env_settings.json(),
            jvm_settings.json(),
            resource_requirement_settings.json(),
        )

        return ray.get(obj_ref1)

    def cluster(self, name: str) -> ray.actor.ActorHandle:
        if not self.launched:
            raise Exception("Please launch gateway first")

        if name in self.clusters:
            return self.clusters[name]

        cluster_ref = self.retrieval_gateway.getCluster.remote(name)
        # master_ref.buildFromRayObjectStore.remote("db1","table1",data_refs)
        cluster = ray.get(cluster_ref)
        self.clusters[name] = cluster
        return cluster

    def cluster_info(self, name: str) -> Dict[str, Any]:
        cluster = self.cluster(name)
        return json.loads(ray.get(cluster.clusterInfo.remote()))

    def is_cluster_exists(self, name: str) -> bool:
        try:
            ray.get_actor(name)
            return True
        except ValueError:
            return False

    def get_table_settings(
        self, cluster_name: str, database: str, table: str
    ) -> Optional[TableSettings]:
        cluster_info = self.cluster_info(cluster_name)
        target_table_settings = None
        for table_settings_dict in cluster_info["tableSettingsList"]:
            table_settings = TableSettings(**table_settings_dict)
            if table_settings.database == database and table_settings.table == table:
                target_table_settings = table_settings
                break
        return target_table_settings

    def check_table_exists(self, cluster_name: str, database: str, table: str) -> bool:
        return self.get_table_settings(cluster_name, database, table) is not None

    def restore_from_cluster_info(self, cluster_info: Dict[str, Any]) -> bool:
        return ray.get(
            self.retrieval_gateway.restoreFromClusterInfo.remote(
                json.dumps(cluster_info, ensure_ascii=False)
            )
        )

    def create_table(self, cluster_name: str, tableSettings: TableSettings) -> bool:

        if self.check_table_exists(
            cluster_name, tableSettings.database, tableSettings.table
        ):
            raise Exception(
                f"Table {tableSettings.database}.{tableSettings.table} already exists in cluster {cluster_name}"
            )

        cluster = self.cluster(cluster_name)
        return ray.get(cluster.createTable.remote(tableSettings.json()))

    def build(
        self,
        cluster_name: str,
        database: str,
        table: str,
        object_refs: List[ObjectRef[str]],
    ) -> bool:

        if not self.check_table_exists(cluster_name, database, table):
            raise Exception(
                f"Table {database}.{table} not exists in cluster {cluster_name}"
            )

        cluster = self.cluster(cluster_name)

        data_ids = ref_utils.get_object_ids(object_refs)
        locations = ref_utils.get_locations(object_refs)
        return ray.get(
            cluster.buildFromRayObjectStore.remote(database, table, data_ids, locations)
        )

    def build_from_dicts(
        self, cluster_name: str, database: str, table: str, data: List[Dict[str, Any]]
    ) -> bool:
        data_refs = []

        for item in data:
            itemref = ray.put(json.dumps(item, ensure_ascii=False))
            data_refs.append(itemref)

        return self.build(cluster_name, database, table, data_refs)

    def delete_by_ids(
        self, cluster_name: str, database: str, table: str, ids: List[Any]
    ) -> bool:

        if not self.check_table_exists(cluster_name, database, table):
            raise Exception(
                f"Table {database}.{table} not exists in cluster {cluster_name}"
            )

        cluster = self.cluster(cluster_name)
        return ray.get(
            cluster.deleteByIds.remote(
                database, table, json.dumps(ids, ensure_ascii=False)
            )
        )

    def get_tables(self, cluster_name: str) -> List[TableSettings]:
        cluster_info = self.cluster_info(cluster_name)
        target_table_settings = []
        for table_settings_dict in cluster_info["tableSettingsList"]:
            target_table_settings.append(TableSettings(**table_settings_dict))
        return target_table_settings

    def get_databases(self, cluster_name: str) -> List[str]:
        table_settings_list = self.get_tables(cluster_name)
        return [x.database for x in table_settings_list]

    def shutdown_cluster(self, cluster_name: str) -> bool:
        if not self.launched:
            raise Exception("Please launch gateway first")

        v = ray.get(self.retrieval_gateway.shutdownCluster.remote(cluster_name))
        if cluster_name in self.clusters:
            del self.clusters[cluster_name]
        return v

    def commit(self, cluster_name: str, database: str, table: str) -> bool:

        if not self.check_table_exists(cluster_name, database, table):
            raise Exception(
                f"Table {database}.{table} not exists in cluster {cluster_name}"
            )

        cluster = self.cluster(cluster_name)
        return ray.get(cluster.commit.remote(database, table))

    def truncate(self, cluster_name: str, database: str, table: str) -> bool:

        if not self.check_table_exists(cluster_name, database, table):
            raise Exception(
                f"Table {database}.{table} not exists in cluster {cluster_name}"
            )

        cluster = self.cluster(cluster_name)
        return ray.get(cluster.truncate.remote(database, table))

    def close(self, cluster_name: str, database: str, table: str) -> bool:

        if not self.check_table_exists(cluster_name, database, table):
            raise Exception(
                f"Table {database}.{table} not exists in cluster {cluster_name}"
            )

        cluster = self.cluster(cluster_name)
        return ray.get(cluster.close.remote(database, table))

    def closeAndDeleteFile(self, cluster_name: str, database: str, table: str) -> bool:

        if not self.check_table_exists(cluster_name, database, table):
            raise Exception(
                f"Table {database}.{table} not exists in cluster {cluster_name}"
            )

        cluster = self.cluster(cluster_name)
        return ray.get(cluster.closeAndDeleteFile.remote(database, table))

    def search_keyword(
        self,
        cluster_name: str,
        database: str,
        table: str,
        filters: Dict[str, Any],
        keyword: str,
        fields: List[str],
        limit: int = 10,
    ) -> List[Dict[str, Any]]:

        search = SearchQuery(
            database=database,
            table=table,
            filters=filters,
            keyword=keyword,
            fields=fields,
            vector=[],
            vectorField=None,
            limit=limit,
        )
        cluster = self.cluster(cluster_name)
        v = cluster.search.remote(f"[{search.json()}]")
        return json.loads(ray.get(v))

    def search_vector(
        self,
        cluster_name: str,
        database: str,
        table: str,
        filters: Dict[str, Any],
        vector: List[float],
        vector_field: str,
        limit: int = 10,
    ) -> List[Dict[str, Any]]:

        search = SearchQuery(
            database=database,
            table=table,
            filters=filters,
            keyword=None,
            fields=[],
            vector=vector,
            vectorField=vector_field,
            limit=limit,
        )
        cluster = self.cluster(cluster_name)
        v = cluster.search.remote(f"[{search.json()}]")
        return json.loads(ray.get(v))

    def search(
        self, cluster_name: str, search_query: Union[List[SearchQuery], SearchQuery]
    ) -> List[Dict[str, Any]]:
        cluster = self.cluster(cluster_name)
        if isinstance(search_query, SearchQuery):
            search_query = [search_query]

        if not search_query:
            raise Exception("search_query is empty")

        v = cluster.search.remote(f"[{','.join([x.json() for x in search_query])}]")
        return json.loads(ray.get(v))

    def filter(
        self, cluster_name: str, search_query: Union[List[SearchQuery], SearchQuery]
    ) -> List[Dict[str, Any]]:
        cluster = self.cluster(cluster_name)
        if isinstance(search_query, SearchQuery):
            search_query = [search_query]
        v = cluster.filter.remote(f"[{','.join([x.json() for x in search_query])}]")
        return json.loads(ray.get(v))

    def delete_by_filter(
        self, cluster_name: str, database: str, table: str, filter: Dict[str, Any]
    ) -> bool:
        cluster = self.cluster(cluster_name)
        return ray.get(
            cluster.deleteByFilter.remote(
                database, table, json.dumps(filter, ensure_ascii=False)
            )
        )


class ByzerRetrieval:

    def __init__(self, pure_client: bool = False):
        self.launched = False
        self.retrieval_proxy: ByzerRetrievalProxy = None        
        self.pure_client = pure_client

    def launch_gateway(self) -> Union[ray.actor.ActorHandle, ByzerRetrievalProxy]:        
        if self.pure_client:
            try:
                self.retrieval_proxy = ray.get_actor("ByzerRetrievalProxy")
            except Exception:                
                pass

            if self.retrieval_proxy:                
                self.launched = True
                return self.retrieval_proxy

            self.retrieval_proxy = (
                ray.remote(ByzerRetrievalProxy)
                .options(name="ByzerRetrievalProxy", lifetime="detached")
                .remote()
            )
            ray.get(self.retrieval_proxy.launch_gateway.remote())
            self.launched = True
            return self.retrieval_proxy
        else:
            self.retrieval_proxy = ByzerRetrievalProxy()
            return self.retrieval_proxy.launch_gateway()

    def gateway(self) -> ray.actor.ActorHandle:
        if not self.pure_client:
            return self.retrieval_proxy.gateway()
        return ray.get(self.retrieval_proxy.gateway.remote())

    def cluster_builder(self) -> ClusterBuilder:
        if not self.pure_client:
            return self.retrieval_proxy.cluster_builder()
        br = self.retrieval_proxy
        return ClusterBuilder(br,pure_client=self.pure_client)

    def start_cluster(
        self,
        cluster_settings: ClusterSettings,
        env_settings: EnvSettings,
        jvm_settings: JVMSettings,
        resource_requirement_settings: ResourceRequirementSettings = ResourceRequirementSettings(
            []
        ),
    ) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.start_cluster(
                cluster_settings,
                env_settings,
                jvm_settings,
                resource_requirement_settings,
            )
        return ray.get(
            self.retrieval_proxy.start_cluster.remote(
                cluster_settings,
                env_settings,
                jvm_settings,
                resource_requirement_settings,
            )
        )

    def cluster(self, name: str) -> ray.actor.ActorHandle:
        if not self.pure_client:
            return self.retrieval_proxy.cluster(name)
        return ray.get(self.retrieval_proxy.cluster.remote(name))

    def cluster_info(self, name: str) -> Dict[str, Any]:
        if not self.pure_client:
            return self.retrieval_proxy.cluster_info(name)
        return ray.get(self.retrieval_proxy.cluster_info.remote(name))

    def is_cluster_exists(self, name: str) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.is_cluster_exists(name)
        return ray.get(self.retrieval_proxy.is_cluster_exists.remote(name))

    def get_table_settings(
        self, cluster_name: str, database: str, table: str
    ) -> Optional[TableSettings]:
        if not self.pure_client:
            return self.retrieval_proxy.get_table_settings(
                cluster_name, database, table
            )
        return ray.get(
            self.retrieval_proxy.get_table_settings.remote(
                cluster_name, database, table
            )
        )

    def check_table_exists(self, cluster_name: str, database: str, table: str) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.check_table_exists(
                cluster_name, database, table
            )
        return ray.get(
            self.retrieval_proxy.check_table_exists.remote(
                cluster_name, database, table
            )
        )

    def restore_from_cluster_info(self, cluster_info: Dict[str, Any]) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.restore_from_cluster_info(cluster_info)
        return ray.get(
            self.retrieval_proxy.restore_from_cluster_info.remote(cluster_info)
        )

    def create_table(self, cluster_name: str, tableSettings: TableSettings) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.create_table(cluster_name, tableSettings)
        return ray.get(
            self.retrieval_proxy.create_table.remote(cluster_name, tableSettings)
        )

    def build(
        self,
        cluster_name: str,
        database: str,
        table: str,
        object_refs: List[ObjectRef[str]],
    ) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.build(
                cluster_name, database, table, object_refs
            )
        return ray.get(
            self.retrieval_proxy.build.remote(
                cluster_name, database, table, object_refs
            )
        )

    def build_from_dicts(
        self, cluster_name: str, database: str, table: str, data: List[Dict[str, Any]]
    ) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.build_from_dicts(
                cluster_name, database, table, data
            )
        return ray.get(
            self.retrieval_proxy.build_from_dicts.remote(
                cluster_name, database, table, data
            )
        )

    def delete_by_ids(
        self, cluster_name: str, database: str, table: str, ids: List[Any]
    ) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.delete_by_ids(
                cluster_name, database, table, ids
            )

        return ray.get(
            self.retrieval_proxy.delete_by_ids.remote(
                cluster_name, database, table, ids
            )
        )

    def get_tables(self, cluster_name: str) -> List[TableSettings]:
        if not self.pure_client:
            return self.retrieval_proxy.get_tables(cluster_name)

        return ray.get(self.retrieval_proxy.get_tables.remote(cluster_name))

    def get_databases(self, cluster_name: str) -> List[str]:
        if not self.pure_client:
            return self.retrieval_proxy.get_databases(cluster_name)

        return ray.get(self.retrieval_proxy.get_databases.remote(cluster_name))

    def shutdown_cluster(self, cluster_name: str) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.shutdown_cluster(cluster_name)

        return ray.get(self.retrieval_proxy.shutdown_cluster.remote(cluster_name))

    def commit(self, cluster_name: str, database: str, table: str) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.commit(cluster_name, database, table)

        return ray.get(
            self.retrieval_proxy.commit.remote(cluster_name, database, table)
        )

    def truncate(self, cluster_name: str, database: str, table: str) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.truncate(cluster_name, database, table)

        return ray.get(
            self.retrieval_proxy.truncate.remote(cluster_name, database, table)
        )

    def close(self, cluster_name: str, database: str, table: str) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.close(cluster_name, database, table)

        return ray.get(self.retrieval_proxy.close.remote(cluster_name, database, table))

    def closeAndDeleteFile(self, cluster_name: str, database: str, table: str) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.closeAndDeleteFile(
                cluster_name, database, table
            )

        return ray.get(
            self.retrieval_proxy.closeAndDeleteFile.remote(
                cluster_name, database, table
            )
        )

    def search_keyword(
        self,
        cluster_name: str,
        database: str,
        table: str,
        filters: Dict[str, Any],
        keyword: str,
        fields: List[str],
        limit: int = 10,
    ) -> List[Dict[str, Any]]:
        if not self.pure_client:
            return self.retrieval_proxy.search_keyword(
                cluster_name, database, table, filters, keyword, fields, limit
            )
        return ray.get(
            self.retrieval_proxy.search_keyword.remote(
                cluster_name, database, table, filters, keyword, fields, limit
            )
        )

    def search_vector(
        self,
        cluster_name: str,
        database: str,
        table: str,
        filters: Dict[str, Any],
        vector: List[float],
        vector_field: str,
        limit: int = 10,
    ) -> List[Dict[str, Any]]:
        if not self.pure_client:
            return self.retrieval_proxy.search_vector(
                cluster_name, database, table, filters, vector, vector_field, limit
            )
        return ray.get(
            self.retrieval_proxy.search_vector.remote(
                cluster_name, database, table, filters, vector, vector_field, limit
            )
        )

    def search(
        self, cluster_name: str, search_query: Union[List[SearchQuery], SearchQuery]
    ) -> List[Dict[str, Any]]:
        if not self.pure_client:
            return self.retrieval_proxy.search(cluster_name, search_query)

        return ray.get(self.retrieval_proxy.search.remote(cluster_name, search_query))

    def filter(
        self, cluster_name: str, search_query: Union[List[SearchQuery], SearchQuery]
    ) -> List[Dict[str, Any]]:
        if not self.pure_client:
            return self.retrieval_proxy.filter(cluster_name, search_query)
        return ray.get(self.retrieval_proxy.filter.remote(cluster_name, search_query))

    def delete_by_filter(
        self, cluster_name: str, database: str, table: str, filter: Dict[str, Any]
    ) -> bool:
        if not self.pure_client:
            return self.retrieval_proxy.delete_by_filter(
                cluster_name, database, table, filter
            )

        return ray.get(
            self.retrieval_proxy.delete_by_filter.remote(
                cluster_name, database, table, filter
            )
        )



import os
import json
from typing import List, Dict
from urllib.parse import urlparse

MODELS_JSON = os.path.expanduser("~/.auto-coder/keys/models.json")

# Default built-in models
default_models_list = [
    {
        "name": "deepseek/r1",
        "description": "DeepSeek Reasoner is for design/review",
        "model_name": "deepseek-reasoner",
        "model_type": "saas/openai",
        "base_url": "https://api.deepseek.com/v1",
        "api_key_path": "api.deepseek.com",
        "is_reasoning": True,
        "input_price": 0.0,  # 单位:M/百万 input tokens
        "output_price": 0.0,  # 单位:M/百万 output tokens
        "average_speed": 0.0,  # 单位:秒/请求
        "max_output_tokens": 8096
    },    
    {
        "name": "deepseek/v3",
        "description": "DeepSeek Chat is for coding",
        "model_name": "deepseek-chat",
        "model_type": "saas/openai",
        "base_url": "https://api.deepseek.com/v1",
        "api_key_path": "api.deepseek.com",
        "is_reasoning": False,
        "input_price": 0.0,
        "output_price": 0.0,
        "average_speed": 0.0,
        "max_output_tokens": 8096
    },
    {
        "name": "ark/deepseek-v3-250324",
        "description": "DeepSeek Chat is for coding",
        "model_name": "deepseek-v3-250324",
        "model_type": "saas/openai",
        "base_url": "https://ark.cn-beijing.volces.com/api/v3",
        "api_key_path": "",
        "is_reasoning": False,
        "input_price": 2.0,
        "output_price": 8.0,
        "average_speed": 0.0,
        "max_output_tokens": 8096
    },    
    {
        "name": "ark/deepseek-r1-250120",
        "description": "DeepSeek Reasoner is for design/review",
        "model_name": "deepseek-r1-250120",
        "model_type": "saas/openai",
        "base_url": "https://ark.cn-beijing.volces.com/api/v3",
        "api_key_path": "",
        "is_reasoning": True,
        "input_price": 4.0,
        "output_price": 16.0,
        "average_speed": 0.0,
        "max_output_tokens": 8096
    },    
    {
        "name": "openai/gpt-4.1-mini",
        "description": "",
        "model_name": "openai/gpt-4.1-mini",
        "model_type": "saas/openai",
        "base_url": "https://openrouter.ai/api/v1",
        "api_key_path": "",
        "is_reasoning": False,
        "input_price": 2.8,
        "output_price": 11.2,
        "average_speed": 0.0,
        "max_output_tokens": 8096*3
    },
    {
        "name": "openai/gpt-4.1",
        "description": "",
        "model_name": "openai/gpt-4.1",
        "model_type": "saas/openai",
        "base_url": "https://openrouter.ai/api/v1",
        "api_key_path": "",
        "is_reasoning": False,
        "input_price": 14.0,
        "output_price": 42.0,
        "average_speed": 0.0,
        "max_output_tokens": 8096*3
    },   
    {
        "name": "openai/gpt-4.1-nano",
        "description": "",
        "model_name": "openai/gpt-4.1-nano",
        "model_type": "saas/openai",
        "base_url": "https://openrouter.ai/api/v1",
        "api_key_path": "",
        "is_reasoning": False,
        "input_price": 0.0,
        "output_price": 0.0,
        "average_speed": 0.0,
        "max_output_tokens": 8096*3
    },  
    {
        "name": "openrouter/google/gemini-2.5-pro-preview-03-25",
        "description": "",
        "model_name": "google/gemini-2.5-pro-preview-03-25",
        "model_type": "saas/openai",
        "base_url": "https://openrouter.ai/api/v1",
        "api_key_path": "",
        "is_reasoning": False,
        "input_price": 0.0,
        "output_price": 0.0,
        "average_speed": 0.0,
        "max_output_tokens": 8096*2
    }
]

def process_api_key_path(base_url: str) -> str:
    """
    从 base_url 中提取 host 部分并处理特殊字符
    例如: https://api.example.com:8080/v1 -> api.example.com_8080
    """
    if not base_url:
        return ""
    
    parsed = urlparse(base_url)
    host = parsed.netloc
    
    # 将冒号替换为下划线
    host = host.replace(":", "_")
    
    return host

def load_models() -> List[Dict]:
    """
    Load models from ~/.auto-coder/keys/models.json and merge with default_models_list.
    Models are merged and deduplicated based on their name field.
    If file doesn't exist or is invalid, use default_models_list.
    """
    os.makedirs(os.path.dirname(MODELS_JSON), exist_ok=True)
    
    # Start with default models
    models_dict = {model["name"]: model for model in default_models_list}
    
    # If JSON file exists, read and merge with defaults
    if os.path.exists(MODELS_JSON):
        try:
            with open(MODELS_JSON, 'r', encoding='utf-8') as f:
                custom_models = json.load(f)
                # Custom models will override defaults with same name
                for model in custom_models:
                    model["is_reasoning"] = model.get("is_reasoning", False)
                    models_dict[model["name"]] = model

        except json.JSONDecodeError:
            # If JSON is invalid, just use defaults
            print("JSON is invalid, using defaults")
            save_models(default_models_list)
    else:
        # If file doesn't exist, create it with defaults
        save_models(default_models_list)
    
    # Convert merged dictionary back to list
    target_models = list(models_dict.values())
    api_key_dir = os.path.expanduser("~/.auto-coder/keys")
    for model in target_models:    
        if model.get("api_key_path",""):           
            api_key_file = os.path.join(api_key_dir, model["api_key_path"])
            if os.path.exists(api_key_file):
                with open(api_key_file, "r",encoding="utf-8") as f:
                    model["api_key"] = f.read().strip()                   
    return target_models

def save_models(models: List[Dict]) -> None:
    """
    Save models to ~/.auto-coder/keys/models.json
    """
    os.makedirs(os.path.dirname(MODELS_JSON), exist_ok=True)
    with open(MODELS_JSON, 'w', encoding='utf-8') as f:
        json.dump(models, f, indent=2, ensure_ascii=False)


def add_and_activate_models(models: List[Dict]) -> None:
    """
    添加模型
    """
    exits_models = load_models()
    for model in models:        
        if model["name"] not in [m["name"] for m in exits_models]:
            exits_models.append(model)
    save_models(exits_models)

    for model in models:
        if "api_key" in model:
            update_model_with_api_key(model["name"], model["api_key"])

def get_model_by_name(name: str) -> Dict:
    """
    根据模型名称查找模型
    """
    from autocoder.common.auto_coder_lang import get_message_with_format
    models = load_models()
    v = [m for m in models if m["name"] == name.strip()]
    
    if len(v) == 0:        
        raise Exception(get_message_with_format("model_not_found", model_name=name))
    return v[0]


def update_model_input_price(name: str, price: float) -> bool:
    """更新模型输入价格
    
    Args:
        name (str): 要更新的模型名称，必须与models.json中的记录匹配
        price (float): 新的输入价格，单位：美元/百万tokens。必须大于等于0
        
    Returns:
        bool: 是否成功找到并更新了模型价格
        
    Raises:
        ValueError: 如果price为负数时抛出
        
    Example:
        >>> update_model_input_price("gpt-4", 3.0)
        True
        
    Notes:
        1. 价格设置后会立即生效并保存到models.json
        2. 实际费用计算时会按实际使用量精确到小数点后6位
        3. 设置价格为0表示该模型当前不可用
    """
    if price < 0:
        raise ValueError("Price cannot be negative")
        
    models = load_models()
    updated = False
    for model in models:
        if model["name"] == name:
            model["input_price"] = float(price)
            updated = True
            break
    if updated:
        save_models(models)
    return updated

def update_model_output_price(name: str, price: float) -> bool:
    """更新模型输出价格
    
    Args:
        name (str): 要更新的模型名称，必须与models.json中的记录匹配
        price (float): 新的输出价格，单位：美元/百万tokens。必须大于等于0
        
    Returns:
        bool: 是否成功找到并更新了模型价格
        
    Raises:
        ValueError: 如果price为负数时抛出
        
    Example:
        >>> update_model_output_price("gpt-4", 6.0)
        True
        
    Notes:
        1. 输出价格通常比输入价格高30%-50%
        2. 对于按token计费的API，实际收费按(input_tokens * input_price + output_tokens * output_price)计算
        3. 价格变更会影响所有依赖模型计费的功能（如成本预测、用量监控等）
    """
    if price < 0:
        raise ValueError("Price cannot be negative")
        
    models = load_models()
    updated = False
    for model in models:
        if model["name"] == name:
            model["output_price"] = float(price)
            updated = True
            break
    if updated:
        save_models(models)
    return updated

def update_model_speed(name: str, speed: float) -> bool:
    """更新模型平均速度
    
    Args:
        name: 模型名称
        speed: 速度(秒/请求)
        
    Returns:
        bool: 是否更新成功
    """            
    models = load_models()
    updated = False
    for model in models:
        if model["name"] == name:
            model["average_speed"] = float(speed)
            updated = True
            break
    if updated:
        save_models(models)
    return updated

def check_model_exists(name: str) -> bool:
    """
    检查模型是否存在
    """
    models = load_models()
    return any(m["name"] == name.strip() for m in models)

def update_model_with_api_key(name: str, api_key: str) -> Dict:
    """
    根据模型名称查找并更新模型的 api_key_path。
    如果找到模型,会根据其 base_url 处理 api_key_path。
    
    Args:
        name: 模型名称
        api_key: API密钥
        
    Returns:
        Dict: 更新后的模型信息,如果未找到则返回None
    """
    models = load_models()
    
    # 在现有模型中查找
    found_model = None
    for model in models:
        if model["name"] == name.strip():
            found_model = model
            break
                    
    if not found_model:
        return None
            
    api_key_path = name.replace("/", "_")  # 替换 / 为 _，保证文件名合法
    if api_key_path:
        found_model["api_key_path"] = api_key_path
        
        # 保存 API 密钥
        api_key_dir = os.path.expanduser("~/.auto-coder/keys")
        os.makedirs(api_key_dir, exist_ok=True)
        api_key_file = os.path.join(api_key_dir, api_key_path)
        with open(api_key_file, "w",encoding="utf-8") as f:
            f.write(api_key.strip())
        
        # 如果是新模型，添加到模型列表中
        if all(model["name"] != name for model in models):
            models.append(found_model)
        else:
            # 更新现有模型
            for i, model in enumerate(models):
                if model["name"] == name:
                    models[i] = found_model
                    break
        
        save_models(models)
    
    return found_model

def update_model(name: str, model_data: Dict) -> Dict:
    """
    更新模型信息
    
    Args:
        name: 要更新的模型名称
        model_data: 包含模型新信息的字典，可以包含以下字段:
            - name: 模型名称
            - description: 模型描述
            - model_name: 模型实际名称
            - model_type: 模型类型
            - base_url: 基础URL
            - api_key: API密钥
            - is_reasoning: 是否为推理模型
            - input_price: 输入价格
            - output_price: 输出价格
            - max_output_tokens: 最大输出tokens
            - average_speed: 平均速度
            
    Returns:
        Dict: 更新后的模型信息，如果未找到则返回None
    """
    models = load_models()
    
    # 查找要更新的模型
    found = False
    for i, model in enumerate(models):
        if model["name"] == name:
            # 更新模型字段
            if "description" in model_data:
                model["description"] = model_data["description"]
            if "model_name" in model_data:
                model["model_name"] = model_data["model_name"]
            if "model_type" in model_data:
                model["model_type"] = model_data["model_type"]
            if "base_url" in model_data:
                model["base_url"] = model_data["base_url"]
            if "is_reasoning" in model_data:
                model["is_reasoning"] = model_data["is_reasoning"]
            if "input_price" in model_data:
                model["input_price"] = float(model_data["input_price"])
            if "output_price" in model_data:
                model["output_price"] = float(model_data["output_price"])
            if "max_output_tokens" in model_data:
                model["max_output_tokens"] = int(model_data["max_output_tokens"])
            if "average_speed" in model_data:
                model["average_speed"] = float(model_data["average_speed"])
            
            # 保存更新后的模型
            models[i] = model
            found = True
            
            # 如果提供了API密钥，则更新
            if "api_key" in model_data and model_data["api_key"]:
                update_model_with_api_key(name, model_data["api_key"])
            
            break
    
    if found:
        save_models(models)
        return models[i]
    
    return None





"""
Media Processors Module - Handle image and audio file processing
"""

import base64
import os
import mimetypes
from pathlib import Path
from typing import Optional, Tuple, Dict, Any
import logging
from PIL import Image as PILImage
import io

logger = logging.getLogger(__name__)


class ProcessorError(Exception):
    """Custom exception for processor errors"""
    pass


class BaseProcessor:
    """Base class for media processors"""
    
    def __init__(self, max_file_size: int = 10 * 1024 * 1024):  # 10MB default
        self.max_file_size = max_file_size
        self.supported_formats = set()
        
    def validate_file_path(self, path: str) -> Path:
        """
        Validate that the file path exists and is accessible.
        
        Args:
            path: File path to validate
            
        Returns:
            Path object
            
        Raises:
            ProcessorError: If file doesn't exist or is inaccessible
        """
        file_path = Path(path)
        
        if not file_path.exists():
            raise ProcessorError(f"File not found: {path}")
            
        if not file_path.is_file():
            raise ProcessorError(f"Path is not a file: {path}")
            
        if not os.access(file_path, os.R_OK):
            raise ProcessorError(f"File is not readable: {path}")
            
        # Check file size
        file_size = file_path.stat().st_size
        if file_size > self.max_file_size:
            raise ProcessorError(
                f"File size ({file_size} bytes) exceeds maximum allowed size ({self.max_file_size} bytes)"
            )
            
        return file_path
    
    def get_mime_type(self, file_path: Path) -> str:
        """Get MIME type of the file"""
        mime_type, _ = mimetypes.guess_type(str(file_path))
        return mime_type or 'application/octet-stream'
    
    def validate_base64(self, data: str) -> bool:
        """Validate base64 string"""
        try:
            # Check if it's a data URI
            if data.startswith('data:'):
                # Extract base64 part
                parts = data.split(',', 1)
                if len(parts) != 2:
                    return False
                data = parts[1]
                
            # Try to decode
            base64.b64decode(data, validate=True)
            return True
        except Exception:
            return False


class ImageProcessor(BaseProcessor):
    """
    Process images: load from path, convert to base64, save from base64
    """
    
    def __init__(self, max_file_size: int = 10 * 1024 * 1024):
        super().__init__(max_file_size)
        self.supported_formats = {
            '.jpg', '.jpeg', '.png', '.gif', '.bmp', 
            '.webp', '.ico', '.tiff', '.svg'
        }
        
    def load_image_from_path(self, 
                           path: str, 
                           validate_image: bool = True,
                           as_data_uri: bool = True) -> str:
        """
        Load image from file path and convert to base64.
        
        Args:
            path: Path to the image file
            validate_image: Whether to validate the image can be opened
            as_data_uri: Return as data URI format (data:image/type;base64,...)
            
        Returns:
            Base64 encoded image string
            
        Raises:
            ProcessorError: If image cannot be loaded or validated
        """
        file_path = self.validate_file_path(path)
        
        # Check file extension
        extension = file_path.suffix.lower()
        if extension not in self.supported_formats:
            logger.warning(f"Unsupported image format: {extension}")
        
        # Validate image if requested
        if validate_image and extension != '.svg':
            try:
                with PILImage.open(file_path) as img:
                    # Verify the image can be loaded
                    img.verify()
            except Exception as e:
                raise ProcessorError(f"Invalid image file: {str(e)}")
        
        # Read and encode the file
        try:
            with open(file_path, 'rb') as f:
                image_data = f.read()
                encoded = base64.b64encode(image_data).decode('utf-8')
                
            if as_data_uri:
                # Determine image type
                mime_type = self.get_mime_type(file_path)
                if mime_type == 'application/octet-stream':
                    # Fallback to extension-based type
                    image_type = extension[1:] if extension else 'png'
                    mime_type = f'image/{image_type}'
                    
                return f"data:{mime_type};base64,{encoded}"
            else:
                return encoded
                
        except Exception as e:
            raise ProcessorError(f"Failed to read image file: {str(e)}")
    
    def save_image_from_base64(self,
                             image_data: str,
                             output_path: str,
                             auto_fix_extension: bool = True,
                             validate_image: bool = True) -> str:
        """
        Save base64 image data to file.
        
        Args:
            image_data: Base64 encoded image (with or without data URI prefix)
            output_path: Path where to save the image
            auto_fix_extension: Automatically fix file extension based on image type
            validate_image: Validate the decoded image before saving
            
        Returns:
            Final output path (may differ if auto_fix_extension is True)
            
        Raises:
            ProcessorError: If image data is invalid or cannot be saved
        """
        # Validate base64 data
        if not self.validate_base64(image_data):
            raise ProcessorError("Invalid base64 image data")
        
        # Parse data URI if present
        if image_data.startswith('data:'):
            try:
                header, encoded = image_data.split(',', 1)
                # Extract MIME type
                mime_type = header.split(':')[1].split(';')[0]
                image_type = mime_type.split('/')[1]
            except Exception:
                raise ProcessorError("Invalid data URI format")
        else:
            encoded = image_data
            image_type = None
        
        # Decode base64
        try:
            decoded_data = base64.b64decode(encoded)
        except Exception as e:
            raise ProcessorError(f"Failed to decode base64 data: {str(e)}")
        
        # Validate image if requested
        if validate_image:
            try:
                img = PILImage.open(io.BytesIO(decoded_data))
                img.verify()
                # Get actual format if not provided
                if image_type is None:
                    image_type = img.format.lower()
            except Exception as e:
                raise ProcessorError(f"Invalid image data: {str(e)}")
        
        # Handle output path and extension
        output_path = Path(output_path)
        
        if auto_fix_extension and image_type:
            correct_extension = f'.{image_type}'
            if output_path.suffix.lower() != correct_extension:
                output_path = output_path.with_suffix(correct_extension)
        
        # Create parent directories if needed
        output_path.parent.mkdir(parents=True, exist_ok=True)
        
        # Write the file
        try:
            with open(output_path, 'wb') as f:
                f.write(decoded_data)
        except Exception as e:
            raise ProcessorError(f"Failed to save image: {str(e)}")
        
        return str(output_path)
    
    def get_image_info(self, path: str) -> Dict[str, Any]:
        """
        Get information about an image file.
        
        Returns:
            Dictionary with image metadata
        """
        file_path = self.validate_file_path(path)
        
        info = {
            'path': str(file_path),
            'size': file_path.stat().st_size,
            'mime_type': self.get_mime_type(file_path)
        }
        
        try:
            with PILImage.open(file_path) as img:
                info.update({
                    'format': img.format,
                    'mode': img.mode,
                    'width': img.width,
                    'height': img.height
                })
        except Exception as e:
            logger.warning(f"Could not read image metadata: {str(e)}")
            
        return info


class AudioProcessor(BaseProcessor):
    """
    Process audio files: load from path, convert to base64, save from base64
    """
    
    def __init__(self, max_file_size: int = 50 * 1024 * 1024):  # 50MB default for audio
        super().__init__(max_file_size)
        self.supported_formats = {
            '.mp3', '.wav', '.ogg', '.m4a', '.flac',
            '.aac', '.wma', '.opus', '.webm'
        }
        
    def load_audio_from_path(self,
                           path: str,
                           as_data_uri: bool = True) -> str:
        """
        Load audio from file path and convert to base64.
        
        Args:
            path: Path to the audio file
            as_data_uri: Return as data URI format
            
        Returns:
            Base64 encoded audio string
        """
        file_path = self.validate_file_path(path)
        
        # Check file extension
        extension = file_path.suffix.lower()
        if extension not in self.supported_formats:
            logger.warning(f"Unsupported audio format: {extension}")
        
        # Read and encode the file
        try:
            with open(file_path, 'rb') as f:
                audio_data = f.read()
                encoded = base64.b64encode(audio_data).decode('utf-8')
                
            if as_data_uri:
                mime_type = self.get_mime_type(file_path)
                if mime_type == 'application/octet-stream':
                    # Fallback to extension-based type
                    audio_type = extension[1:] if extension else 'mpeg'
                    mime_type = f'audio/{audio_type}'
                    
                return f"data:{mime_type};base64,{encoded}"
            else:
                return encoded
                
        except Exception as e:
            raise ProcessorError(f"Failed to read audio file: {str(e)}")
    
    def save_audio_from_base64(self,
                             audio_data: str,
                             output_path: str,
                             auto_fix_extension: bool = True) -> str:
        """
        Save base64 audio data to file.
        
        Args:
            audio_data: Base64 encoded audio
            output_path: Path where to save the audio
            auto_fix_extension: Automatically fix file extension
            
        Returns:
            Final output path
        """
        # Validate base64 data
        if not self.validate_base64(audio_data):
            raise ProcessorError("Invalid base64 audio data")
        
        # Parse data URI if present
        if audio_data.startswith('data:'):
            try:
                header, encoded = audio_data.split(',', 1)
                mime_type = header.split(':')[1].split(';')[0]
                audio_type = mime_type.split('/')[1]
            except Exception:
                raise ProcessorError("Invalid data URI format")
        else:
            encoded = audio_data
            audio_type = None
        
        # Decode base64
        try:
            decoded_data = base64.b64decode(encoded)
        except Exception as e:
            raise ProcessorError(f"Failed to decode base64 data: {str(e)}")
        
        # Handle output path and extension
        output_path = Path(output_path)
        
        if auto_fix_extension and audio_type:
            correct_extension = f'.{audio_type}'
            if output_path.suffix.lower() != correct_extension:
                output_path = output_path.with_suffix(correct_extension)
        
        # Create parent directories if needed
        output_path.parent.mkdir(parents=True, exist_ok=True)
        
        # Write the file
        try:
            with open(output_path, 'wb') as f:
                f.write(decoded_data)
        except Exception as e:
            raise ProcessorError(f"Failed to save audio: {str(e)}")
        
        return str(output_path)






"""
Unit tests for the media processors module
"""

import pytest
import base64
import tempfile
import os
from pathlib import Path
from PIL import Image
import io

from ..processors import (
    ImageProcessor, AudioProcessor, ProcessorError, BaseProcessor
)


class TestBaseProcessor:
    """Test cases for BaseProcessor"""
    
    def test_validate_file_path_valid(self, tmp_path):
        """Test validating a valid file path"""
        # Create a test file
        test_file = tmp_path / "test.txt"
        test_file.write_text("test content")
        
        processor = BaseProcessor()
        result = processor.validate_file_path(str(test_file))
        
        assert isinstance(result, Path)
        assert result.exists()
    
    def test_validate_file_path_not_exists(self):
        """Test validating non-existent file"""
        processor = BaseProcessor()
        
        with pytest.raises(ProcessorError) as exc_info:
            processor.validate_file_path("/path/that/does/not/exist.txt")
        assert "File not found" in str(exc_info.value)
    
    def test_validate_file_path_is_directory(self, tmp_path):
        """Test validating a directory instead of file"""
        processor = BaseProcessor()
        
        with pytest.raises(ProcessorError) as exc_info:
            processor.validate_file_path(str(tmp_path))
        assert "Path is not a file" in str(exc_info.value)
    
    def test_validate_file_size_limit(self, tmp_path):
        """Test file size validation"""
        # Create a file that's too large
        test_file = tmp_path / "large.txt"
        test_file.write_bytes(b"x" * 1000)
        
        processor = BaseProcessor(max_file_size=500)  # 500 bytes limit
        
        with pytest.raises(ProcessorError) as exc_info:
            processor.validate_file_path(str(test_file))
        assert "exceeds maximum allowed size" in str(exc_info.value)
    
    def test_get_mime_type(self, tmp_path):
        """Test MIME type detection"""
        processor = BaseProcessor()
        
        # Test various file types
        test_cases = [
            ("test.jpg", "image/jpeg"),
            ("test.png", "image/png"),
            ("test.mp3", "audio/mpeg"),
            ("test.txt", "text/plain"),
            ("test.unknown", "application/octet-stream")
        ]
        
        for filename, expected_mime in test_cases:
            test_file = tmp_path / filename
            test_file.write_text("dummy")
            
            mime_type = processor.get_mime_type(test_file)
            assert mime_type == expected_mime
    
    def test_validate_base64_valid(self):
        """Test validating valid base64 strings"""
        processor = BaseProcessor()
        
        # Test plain base64
        valid_b64 = base64.b64encode(b"test data").decode()
        assert processor.validate_base64(valid_b64) is True
        
        # Test data URI
        valid_data_uri = f"data:image/png;base64,{valid_b64}"
        assert processor.validate_base64(valid_data_uri) is True
    
    def test_validate_base64_invalid(self):
        """Test validating invalid base64 strings"""
        processor = BaseProcessor()
        
        # Invalid base64
        assert processor.validate_base64("not valid base64!@#") is False
        
        # Invalid data URI
        assert processor.validate_base64("data:image/png;base64,invalid!") is False
        
        # Malformed data URI
        assert processor.validate_base64("data:image/png") is False


class TestImageProcessor:
    """Test cases for ImageProcessor"""
    
    @pytest.fixture
    def sample_image(self, tmp_path):
        """Create a sample image for testing"""
        img_path = tmp_path / "test_image.png"
        
        # Create a simple 10x10 red image
        img = Image.new('RGB', (10, 10), color='red')
        img.save(img_path)
        
        return img_path
    
    def test_load_image_from_path(self, sample_image):
        """Test loading image and converting to base64"""
        processor = ImageProcessor()
        
        # Test with data URI
        result = processor.load_image_from_path(str(sample_image))
        assert result.startswith("data:image/png;base64,")
        
        # Test without data URI
        result = processor.load_image_from_path(str(sample_image), as_data_uri=False)
        assert not result.startswith("data:")
        # Verify it's valid base64
        base64.b64decode(result)
    
    def test_load_image_validation(self, tmp_path):
        """Test image validation during loading"""
        # Create an invalid image file
        invalid_img = tmp_path / "invalid.png"
        invalid_img.write_bytes(b"not an image")
        
        processor = ImageProcessor()
        
        with pytest.raises(ProcessorError) as exc_info:
            processor.load_image_from_path(str(invalid_img), validate_image=True)
        assert "Invalid image file" in str(exc_info.value)
    
    def test_load_image_unsupported_format_warning(self, tmp_path, caplog):
        """Test warning for unsupported formats"""
        # Create a file with unsupported extension
        weird_file = tmp_path / "image.xyz"
        # Write valid PNG data but with weird extension
        img = Image.new('RGB', (10, 10))
        img.save(weird_file, format='PNG')
        
        processor = ImageProcessor()
        processor.load_image_from_path(str(weird_file), validate_image=False)
        
        assert "Unsupported image format: .xyz" in caplog.text
    
    def test_save_image_from_base64(self, sample_image, tmp_path):
        """Test saving image from base64 data"""
        processor = ImageProcessor()
        
        # Load image as base64
        base64_data = processor.load_image_from_path(str(sample_image))
        
        # Save to new location
        output_path = tmp_path / "output" / "saved_image.png"
        saved_path = processor.save_image_from_base64(base64_data, str(output_path))
        
        assert Path(saved_path).exists()
        assert Path(saved_path).suffix == ".png"
        
        # Verify the saved image is valid
        img = Image.open(saved_path)
        assert img.size == (10, 10)
    
    def test_save_image_auto_fix_extension(self, sample_image, tmp_path):
        """Test automatic extension fixing"""
        processor = ImageProcessor()
        
        # Load PNG image
        base64_data = processor.load_image_from_path(str(sample_image))
        
        # Try to save with wrong extension
        output_path = tmp_path / "image.jpg"
        saved_path = processor.save_image_from_base64(
            base64_data, 
            str(output_path),
            auto_fix_extension=True
        )
        
        # Should be saved as .png
        assert Path(saved_path).suffix == ".png"
    
    def test_save_image_invalid_base64(self, tmp_path):
        """Test saving with invalid base64 data"""
        processor = ImageProcessor()
        
        with pytest.raises(ProcessorError) as exc_info:
            processor.save_image_from_base64(
                "invalid base64 data",
                str(tmp_path / "output.png")
            )
        assert "Invalid base64 image data" in str(exc_info.value)
    
    def test_get_image_info(self, sample_image):
        """Test getting image information"""
        processor = ImageProcessor()
        
        info = processor.get_image_info(str(sample_image))
        
        assert info['path'] == str(sample_image)
        assert info['size'] > 0
        assert info['mime_type'] == 'image/png'
        assert info['format'] == 'PNG'
        assert info['width'] == 10
        assert info['height'] == 10
        assert 'mode' in info
    
    def test_supported_formats(self):
        """Test supported image formats"""
        processor = ImageProcessor()
        
        expected_formats = {
            '.jpg', '.jpeg', '.png', '.gif', '.bmp',
            '.webp', '.ico', '.tiff', '.svg'
        }
        
        assert processor.supported_formats == expected_formats


class TestAudioProcessor:
    """Test cases for AudioProcessor"""
    
    @pytest.fixture
    def sample_audio(self, tmp_path):
        """Create a sample audio file for testing"""
        # Create a dummy audio file (just bytes, not real audio)
        audio_path = tmp_path / "test_audio.mp3"
        audio_path.write_bytes(b"dummy audio content" * 100)
        return audio_path
    
    def test_load_audio_from_path(self, sample_audio):
        """Test loading audio and converting to base64"""
        processor = AudioProcessor()
        
        # Test with data URI
        result = processor.load_audio_from_path(str(sample_audio))
        assert result.startswith("data:audio/mpeg;base64,")
        
        # Test without data URI
        result = processor.load_audio_from_path(str(sample_audio), as_data_uri=False)
        assert not result.startswith("data:")
        # Verify it's valid base64
        base64.b64decode(result)
    
    def test_save_audio_from_base64(self, sample_audio, tmp_path):
        """Test saving audio from base64 data"""
        processor = AudioProcessor()
        
        # Load audio as base64
        base64_data = processor.load_audio_from_path(str(sample_audio))
        
        # Save to new location
        output_path = tmp_path / "output" / "saved_audio.mp3"
        saved_path = processor.save_audio_from_base64(base64_data, str(output_path))
        
        assert Path(saved_path).exists()
        # mimetypes might return .mpeg for mp3 files
        assert Path(saved_path).suffix in [".mp3", ".mpeg"]
        
        # Verify content matches
        with open(sample_audio, 'rb') as f1, open(saved_path, 'rb') as f2:
            assert f1.read() == f2.read()
    
    def test_audio_max_file_size(self):
        """Test audio processor has larger default file size limit"""
        processor = AudioProcessor()
        assert processor.max_file_size == 50 * 1024 * 1024  # 50MB
    
    def test_supported_audio_formats(self):
        """Test supported audio formats"""
        processor = AudioProcessor()
        
        expected_formats = {
            '.mp3', '.wav', '.ogg', '.m4a', '.flac',
            '.aac', '.wma', '.opus', '.webm'
        }
        
        assert processor.supported_formats == expected_formats
    
    def test_audio_mime_type_fallback(self, tmp_path):
        """Test MIME type fallback for audio files"""
        # Create audio file with uncommon extension
        audio_file = tmp_path / "audio.m4a"
        audio_file.write_bytes(b"dummy m4a content")
        
        processor = AudioProcessor()
        result = processor.load_audio_from_path(str(audio_file))
        
        # Should have correct MIME type
        assert "data:audio/" in result





"""
Configuration management functionality.

This module provides configuration management methods including
basic configuration, batch operations, and nested configuration support.
"""

from typing import Dict, Any, List
from .base_manager import BaseMemoryManager


class ConfigManagerMixin:
    """Mixin class providing configuration management functionality."""
    
    def get_config(self, key: str, default: Any = None) -> Any:
        """Get configuration value."""
        return self._memory.conf.get(key, default)
    
    def set_config(self, key: str, value: Any):
        """Set configuration value."""
        self._memory.conf[key] = value
        self.save_memory()
    
    def delete_config(self, key: str) -> bool:
        """Delete configuration value."""
        if key in self._memory.conf:
            del self._memory.conf[key]
            self.save_memory()
            return True
        return False
    
    def has_config(self, key: str) -> bool:
        """Check if configuration key exists."""
        return key in self._memory.conf
    
    # Extended configuration management methods
    def get_all_config(self) -> Dict[str, Any]:
        """Get all configuration as a dictionary."""
        return self._memory.conf.copy()
    
    def set_configs(self, config_dict: Dict[str, Any]):
        """Set multiple configuration values at once."""
        for key, value in config_dict.items():
            self._memory.conf[key] = value
        self.save_memory()
    
    def update_config(self, config_dict: Dict[str, Any]):
        """Update configuration with new values (alias for set_configs)."""
        self.set_configs(config_dict)
    
    def clear_config(self):
        """Clear all configuration."""
        self._memory.conf.clear()
        self.save_memory()
    
    def get_config_keys(self) -> List[str]:
        """Get list of all configuration keys."""
        return list(self._memory.conf.keys())
    
    def get_config_count(self) -> int:
        """Get number of configuration items."""
        return len(self._memory.conf)
    
    # Nested configuration support (for keys like "model.name")
    def get_nested_config(self, key: str, default: Any = None) -> Any:
        """
        Get nested configuration value using dot notation.
        
        Args:
            key: Nested key like "model.name" or "database.host"
            default: Default value if key not found
            
        Returns:
            Configuration value or default
        """
        keys = key.split('.')
        value = self._memory.conf
        
        try:
            for k in keys:
                value = value[k]
            return value
        except (KeyError, TypeError):
            return default
    
    def set_nested_config(self, key: str, value: Any):
        """
        Set nested configuration value using dot notation.
        
        Args:
            key: Nested key like "model.name" or "database.host"
            value: Value to set
        """
        keys = key.split('.')
        config = self._memory.conf
        
        # Navigate to the parent of the final key
        for k in keys[:-1]:
            if k not in config:
                config[k] = {}
            elif not isinstance(config[k], dict):
                # If existing value is not a dict, replace it
                config[k] = {}
            config = config[k]
        
        # Set the final value
        config[keys[-1]] = value
        self.save_memory()
    
    def delete_nested_config(self, key: str) -> bool:
        """
        Delete nested configuration value using dot notation.
        
        Args:
            key: Nested key like "model.name" or "database.host"
            
        Returns:
            True if key was found and deleted, False otherwise
        """
        keys = key.split('.')
        config = self._memory.conf
        
        try:
            # Navigate to the parent of the final key
            for k in keys[:-1]:
                config = config[k]
            
            # Delete the final key
            if keys[-1] in config:
                del config[keys[-1]]
                self.save_memory()
                return True
            else:
                return False
        except (KeyError, TypeError):
            return False
    
    def has_nested_config(self, key: str) -> bool:
        """
        Check if nested configuration key exists using dot notation.
        
        Args:
            key: Nested key like "model.name" or "database.host"
            
        Returns:
            True if key exists, False otherwise
        """
        keys = key.split('.')
        config = self._memory.conf
        
        try:
            for k in keys:
                config = config[k]
            return True
        except (KeyError, TypeError):
            return False



"""
Core data models for memory management.

This module contains the data structures used by the memory management system.
"""

from typing import Dict, Any, List
from dataclasses import dataclass, field


@dataclass
class CoreMemory:
    """Core memory data structure for auto-coder sessions."""
    
    conversation: List[Dict[str, str]] = field(default_factory=list)
    current_files: Dict[str, Any] = field(default_factory=lambda: {
        "files": [], 
        "groups": {}, 
        "groups_info": {}, 
        "current_groups": []
    })
    conf: Dict[str, Any] = field(default_factory=dict)
    exclude_dirs: List[str] = field(default_factory=list)
    exclude_files: List[str] = field(default_factory=list)
    libs: Dict[str, Any] = field(default_factory=dict)
    mode: str = "auto_detect"
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary for JSON serialization."""
        return {
            "conversation": self.conversation,
            "current_files": self.current_files,
            "conf": self.conf,
            "exclude_dirs": self.exclude_dirs,
            "exclude_files": self.exclude_files,
            "libs": self.libs,
            "mode": self.mode,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'CoreMemory':
        """Create CoreMemory instance from dictionary data."""
        # Ensure current_files has all required keys
        current_files = data.get("current_files", {})
        if not isinstance(current_files, dict):
            current_files = {"files": [], "groups": {}, "groups_info": {}, "current_groups": []}
        else:
            current_files.setdefault("files", [])
            current_files.setdefault("groups", {})
            current_files.setdefault("groups_info", {})
            current_files.setdefault("current_groups", [])
        
        return cls(
            conversation=data.get("conversation", []),
            current_files=current_files,
            conf=data.get("conf", {}),
            exclude_dirs=data.get("exclude_dirs", []),
            exclude_files=data.get("exclude_files", []),
            libs=data.get("libs", {}),
            mode=data.get("mode", "auto_detect"),
        )


# Type aliases for better code clarity
ConfigDict = Dict[str, Any]
FileList = List[str]
GroupsDict = Dict[str, List[str]]
GroupsInfoDict = Dict[str, Dict[str, Any]]
LibsDict = Dict[str, Any]
ConversationList = List[Dict[str, str]]



import anyio
from concurrent.futures import ThreadPoolExecutor, TimeoutError, CancelledError
from threading import Event
from inspect import signature
from contextlib import contextmanager, closing
from contextlib2 import nullcontext
from functools import wraps, partial
import inspect
import os
import socket
from typing import Any, Optional
import re
from urllib.parse import urlparse
import warnings
import threading
import logging
import time

import json
from typing import Any, Dict, List, Union

@contextmanager
def switch_cwd(path):
    old_cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(old_cwd)

@contextmanager
def patch(obj, attr, val):
    old_val = getattr(obj, attr)
    try:
        setattr(obj, attr, val)
        yield
    finally:
        setattr(obj, attr, old_val)


def asyncfy(func):
    """Decorator that makes a function async. Note that this does not actually make
    the function asynchroniously running in a separate thread, it just wraps it in
    an async function. If you want to actually run the function in a separate thread,
    consider using asyncfy_with_semaphore.

    Args:
        func (function): Function to make async
    """

    if inspect.iscoroutinefunction(func):
        return func

    @wraps(func)
    async def async_func(*args, **kwargs):
        return func(*args, **kwargs)

    return async_func


def asyncfy_with_semaphore(
    func, semaphore: Optional[anyio.Semaphore]=None, timeout: Optional[float] = None
):
    """Decorator that makes a function async, as well as running in a separate thread,
    with the concurrency controlled by the semaphore. If Semaphore is None, we do not
    enforce an upper bound on the number of concurrent calls (but it is still bound by
    the number of threads that anyio defines as an upper bound).

    Args:
        func (function): Function to make async. If the function is already async,
            this function will add semaphore and timeout control to it.
        semaphore (anyio.Semaphore or None): Semaphore to use for concurrency control.
            Concurrent calls to this function will be bounded by the semaphore.
        timeout (float or None): Timeout in seconds. If the function does not return
            within the timeout, a TimeoutError will be raised. If None, no timeout
            will be enforced. If the function is async, one can catch the CancelledError
            inside the function to handle the timeout.
    """
    if inspect.iscoroutinefunction(func):

        @wraps(func)
        async def async_func(*args, **kwargs):
            semaphore_ctx = semaphore if semaphore is not None else nullcontext()
            timeout_ctx = anyio.fail_after(timeout) if timeout else nullcontext()
            with timeout_ctx:
                async with semaphore_ctx:
                    return await func(*args, **kwargs)

        return async_func

    else:

        @wraps(func)
        async def async_func(*args, **kwargs):
            semaphore_ctx = semaphore if semaphore is not None else nullcontext()
            timeout_ctx = anyio.fail_after(timeout) if timeout else nullcontext()
            with timeout_ctx:
                async with semaphore_ctx:
                    return await anyio.to_thread.run_sync(
                        partial(func, *args, **kwargs), cancellable=True
                    )

        return async_func


def is_valid_url(candidate_str: Any) -> bool:
    if not isinstance(candidate_str, str):
        return False
    parsed = urlparse(candidate_str)
    return parsed.scheme != "" and parsed.netloc != ""


# backward compatible function name
def _is_valid_url(candidate_str: Any) -> bool:
    warnings.warn("_is_valid_url is deprecated. Please use is_valid_url instead.")
    return is_valid_url(candidate_str)


def _is_local_url(candidate_str: str) -> bool:
    parsed = urlparse(candidate_str)
    local_hosts = ["localhost", "127.0.0.1", "0.0.0.0", "::1"]
    return parsed.hostname in local_hosts


def find_available_port(port=None):
    if port is None:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    def is_port_occupied(port):
        """
        Returns True if the port is occupied, False otherwise.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(("localhost", port)) == 0

    while is_port_occupied(port):
        print(
            f"Port [yellow]{port}[/] already in use. Incrementing port number to",
            " find an available one."
        )
        port += 1
    return port

def to_bool(s: str) -> bool:
    """
    Convert a string to a boolean value.
    """
    if not isinstance(s, str):
        raise TypeError(f"Expected a string, got {type(s)}")
    true_values = ("yes", "true", "t", "1", "y", "on", "aye", "yea")
    false_values = ("no", "false", "f", "0", "n", "off", "nay", "")
    s = s.lower()
    if s in true_values:
        return True
    elif s in false_values:
        return False
    else:
        raise ValueError(
            f"Invalid boolean value: {s}. Valid true values: {true_values}. Valid false"
            f" values: {false_values}."
        )    

class CancellationRequested(Exception):
    """Raised when a task is requested to be cancelled."""
    pass


def run_in_thread(timeout: Optional[float] = None):
    """Decorator that runs a function in a thread with signal handling.
    
    Args:
        timeout (float, optional): Maximum time to wait for thread completion in seconds.
            If None, will wait indefinitely.
            
    The decorated function will run in a separate thread and can be interrupted by
    signals like Ctrl+C (KeyboardInterrupt). When interrupted, it will log the event
    and clean up gracefully.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(func, *args, **kwargs)
                start_time = time.time()
                
                while True:
                    try:
                        # 使用较短的超时时间进行轮询，确保能够响应中断信号
                        poll_timeout = 0.1
                        if timeout is not None:
                            remaining = timeout - (time.time() - start_time)
                            if remaining <= 0:
                                future.cancel()
                                raise TimeoutError(f"Timeout after {timeout}s in {func.__name__}")
                            poll_timeout = min(poll_timeout, remaining)
                            
                        try:
                            return future.result(timeout=poll_timeout)
                        except TimeoutError:
                            continue  # 继续轮询
                            
                    except KeyboardInterrupt:
                        logging.warning("KeyboardInterrupt received, attempting to cancel task...")
                        future.cancel()
                        raise
                    except Exception as e:
                        logging.error(f"Error occurred in thread: {str(e)}")
                        raise
        return wrapper
    return decorator

def run_in_thread_with_cancel(timeout: Optional[float] = None):
    """Decorator that runs a function in a thread with explicit cancellation support.
    
    Args:
        timeout (float, optional): Maximum time to wait for thread completion in seconds.
            If None, will wait indefinitely.
            
    The decorated function MUST accept 'cancel_event' as its first parameter.
    This cancel_event is a threading.Event object that can be used to check if
    cancellation has been requested.
    
    The decorated function can be called with an external cancel_event passed as a keyword argument.
    If not provided, a new Event will be created.
    
    Example:
        @run_in_thread_with_cancel(timeout=10)
        def long_task(cancel_event, arg1, arg2):
            while not cancel_event.is_set():
                # do work
                if cancel_event.is_set():
                    raise CancellationRequested()
                    
        # 使用外部传入的cancel_event
        external_cancel = Event()
        try:
            result = long_task(arg1, arg2, cancel_event=external_cancel)
        except CancelledError:
            print("Task was cancelled")
            
        # 在其他地方取消任务
        external_cancel.set()
    """
    def decorator(func):
        # 检查函数签名
        sig = signature(func)
        params = list(sig.parameters.keys())
        if not params or params[0] != 'cancel_event':
            raise ValueError(
                f"Function {func.__name__} must have 'cancel_event' as its first parameter. "
                f"Current parameters: {params}"
            )
        
        @wraps(func)
        def wrapper(*args, **kwargs):
            # 从kwargs中提取或创建cancel_event
            cancel_event = kwargs.pop('cancel_event', None) or Event()
            
            def cancellable_task():
                try:
                    return func(cancel_event, *args, **kwargs)
                except CancellationRequested:
                    logging.info(f"Task {func.__name__} was cancelled")
                    raise
                except Exception as e:
                    logging.error(f"Error in {func.__name__}: {str(e)}")
                    raise
            
            with ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(cancellable_task)
                start_time = time.time()
                
                while True:
                    try:
                        # 使用较短的超时时间进行轮询，确保能够响应中断信号
                        poll_timeout = 0.1
                        if timeout is not None:
                            remaining = timeout - (time.time() - start_time)
                            if remaining <= 0:
                                cancel_event.set()
                                future.cancel()
                                raise TimeoutError(f"Timeout after {timeout}s in {func.__name__}")
                            poll_timeout = min(poll_timeout, remaining)
                            
                        try:
                            return future.result(timeout=poll_timeout)
                        except TimeoutError:
                            continue  # 继续轮询
                            
                    except KeyboardInterrupt:
                        logging.warning(f"KeyboardInterrupt received, cancelling {func.__name__}...")
                        cancel_event.set()
                        future.cancel()
                        raise CancelledError("Task cancelled by user")
                    except CancellationRequested:
                        logging.info(f"Task {func.__name__} was cancelled")
                        raise CancelledError("Task cancelled by request")
                    except Exception as e:
                        logging.error(f"Error occurred in thread: {str(e)}")
                        raise
                        
        return wrapper
    return decorator


def run_in_raw_thread():
    """A decorator that runs a function in a separate thread and handles exceptions.
    
    Args:
        func: The function to run in a thread
        
    Returns:
        A wrapper function that executes the decorated function in a thread
        
    The decorator will:
    1. Run the function in a separate thread
    2. Handle KeyboardInterrupt properly
    3. Propagate exceptions from the thread
    4. Support function arguments
    5. Preserve function metadata
    """
    def decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            # Store thread results
            result = []
            exception = []
            
            def worker():            
                ret = func(*args, **kwargs)
                result.append(ret)
            
            # Create and start thread with a meaningful name
            thread = threading.Thread(target=worker, name=f"{func.__name__}_thread")
            thread.daemon = True  # Make thread daemon so it doesn't prevent program exit
            
            try:
                thread.start()
                while thread.is_alive():
                    thread.join(0.1)

                return result[0] if result else None            
            except KeyboardInterrupt:                                
                raise KeyboardInterrupt("Task was cancelled by user")
                
        return wrapper
    return decorator    



from typing import List, Dict, Any, Union, Optional
import pydantic
import base64
import os

# Import from the new modular implementation
from ..nontext import MediaExtractor, TagParser, ImageProcessor, AudioProcessor
from ..nontext.parser import Tag as ModularTag


class Tag(pydantic.BaseModel):
    start_tag: str
    end_tag: str
    content: Union[str, List["Tag"], "Tag"]
    parent: Optional["Tag"] = None


class TagExtractor:
    def __init__(self, text: str):
        self.text = text
        self.pos = -1
        self.len = len(text)
        self.root_tag = Tag(start_tag="<_ROOT_>", end_tag="</_ROOT_>", content=[])
        self.current_tag = None
        self.is_extracted = False

        # Use the new modular parser internally
        self._parser = TagParser()    

    def extract(self) -> Tag:
        if self.is_extracted:
            return self.root_tag
            
        try:
            # Use the new modular parser to parse tags
            parsed_tags = self._parser.parse(self.text)
            
            # Convert the parsed tags back to the old Tag structure
            self._build_tag_tree(parsed_tags)
            
        except Exception:
            # Fallback: if parsing fails, just return root with text content
            self.root_tag.content = self.text
            
        self.is_extracted = True
        return self.root_tag
    
    def _build_tag_tree(self, parsed_tags):
        """Convert modular parser results to old Tag structure"""
        if not parsed_tags:
            self.root_tag.content = self.text
            return
            
        # Sort tags by position
        sorted_tags = sorted(parsed_tags, key=lambda t: t.start_pos)
        
        # Build content list with text and tags
        content_items = []
        last_end = 0
        
        for parsed_tag in sorted_tags:
            # Add text before this tag
            if parsed_tag.start_pos > last_end:
                text_content = self.text[last_end:parsed_tag.start_pos]
                if text_content.strip():
                    content_items.append(text_content)
            
            # Create Tag object for this parsed tag
            tag = Tag(
                start_tag=f"<{parsed_tag.name}>",
                end_tag=f"</{parsed_tag.name}>", 
                content=parsed_tag.content,
                parent=self.root_tag
            )
            content_items.append(tag)
            
            last_end = parsed_tag.end_pos
        
        # Add remaining text
        if last_end < len(self.text):
            remaining_text = self.text[last_end:]
            if remaining_text.strip():
                content_items.append(remaining_text)
        
        # Set the content
        if len(content_items) == 1:
            self.root_tag.content = content_items[0]
        else:
            self.root_tag.content = content_items


class Image(TagExtractor):

    def __init__(self, text: str):
        super().__init__(text)
        self.is_extracted = False
        # Use new modular components
        self._extractor = MediaExtractor()
        self._image_processor = ImageProcessor()

    def has_image(self) -> bool:
        try:
            images = self._extractor.extract_images(self.text, to_base64=False)
            return len(images) > 0
        except:
            # Fallback to original logic
            self.extract()
            for item in self.root_tag.content:
                if (
                    isinstance(item, Tag)
                    and item.start_tag == "<_image_>"
                    and item.end_tag == "</_image_>"
                ):
                    return True
            return False

    @staticmethod
    def convert_image_paths_from(
        text: str, start_tag: str = "<img>", end_tag: str = "</img>"
    ) -> str:
        extractor = MediaExtractor()
        return extractor.convert_tag_format(text, start_tag, "<_image_>", media_type='image')

    @staticmethod
    def extract_image_paths(text: str, to_base64: bool = False) -> List[str]:
        extractor = MediaExtractor()
        # Ensure we get List[str] by setting return_positions=False
        result = extractor.extract_images(text, to_base64=to_base64, return_positions=False)
        # When return_positions=False, result should be List[str]
        return result  # type: ignore

    @staticmethod
    def load_image_from_path(path: str, only_content: bool = False) -> str:
        processor = ImageProcessor()
        result = processor.load_image_from_path(path, as_data_uri=True)
        
        if only_content:
            return result
        return f"<_image_>{result}</_image_>"

    @staticmethod
    def save_image_to_path(
        image_data: str, output_path: str, auto_fix_suffix: bool = False
    ) -> str:
        processor = ImageProcessor()
        return processor.save_image_from_base64(
            image_data, output_path, auto_fix_extension=auto_fix_suffix
        )

    def to_content(self) -> List[Dict[str, str]]:
        try:
            # Use new modular implementation
            return self._extractor.to_content_list(self.text)
        except:
            # Fallback to original logic
            self.extract()

        result = []
        current_item = {}

        for item in self.root_tag.content:
            if isinstance(item, Tag) and item.start_tag == "<_image_>":
                if current_item:
                    result.append(current_item)
                    current_item = {}
                current_item["image"] = item.content

        if current_item:
            result.append(current_item)
            
        new_text = self.text
        for res in result:
            new_text = new_text.replace(f"<_image_>{res['image']}</_image_>", "")

        result.append({"text": new_text.strip()})

        for res in result:
            if "image" in res and not res["image"].startswith("data:image/"):
                # Validate that the content looks like a file path
                image_path = res["image"].strip()
                if self._is_valid_file_path(image_path):
                    try:
                        res["image"] = Image.load_image_from_path(
                            image_path, only_content=True
                        )
                    except Exception:
                        # If loading fails, keep the original path
                        pass

        return result

    def _is_valid_file_path(self, path: str) -> bool:
        """Check if the file path exists"""
        try:
            return os.path.exists(path.strip())
        except:
            return False


class Audio(TagExtractor):
    def __init__(self, text: str):
        super().__init__(text)
        self.is_extracted = False
        # Use new modular components
        self._extractor = MediaExtractor()
        self._audio_processor = AudioProcessor()

    def has_audio(self) -> bool:
        try:
            audio = self._extractor.extract_audio(self.text, to_base64=False)
            return len(audio) > 0
        except:
            # Fallback to original logic
            self.extract()
        for item in self.root_tag.content:
            if (
                isinstance(item, Tag)
                and item.start_tag == "<_audio_>"
                and item.end_tag == "</_audio_>"
            ):
                return True
        return False

    @staticmethod
    def extract_audio_paths(text: str, to_base64: bool = False) -> List[str]:
        extractor = MediaExtractor()
        # Ensure we get List[str] by setting return_positions=False
        result = extractor.extract_audio(text, to_base64=to_base64, return_positions=False)
        # When return_positions=False, result should be List[str]
        return result  # type: ignore

    @staticmethod
    def load_audio_from_path(path: str, only_content: bool = False) -> str:
        processor = AudioProcessor()
        result = processor.load_audio_from_path(path, as_data_uri=True)

        if only_content:
            return result
        return f"<_audio_>{result}</_audio_>"

    @staticmethod
    def save_audio_to_path(
        audio_data: str, output_path: str, auto_fix_suffix: bool = False
    ) -> str:
        processor = AudioProcessor()
        return processor.save_audio_from_base64(
            audio_data, output_path, auto_fix_extension=auto_fix_suffix
        )

    def to_content(self) -> List[Dict[str, str]]:
        try:
            # Use new modular implementation
            return self._extractor.to_content_list(self.text)
        except:
            # Fallback to original logic  
            self.extract()

        result = []
        current_item = {}

        for item in self.root_tag.content:
            if isinstance(item, Tag) and item.start_tag == "<_audio_>":
                if current_item:
                    result.append(current_item)
                    current_item = {}
                current_item["audio"] = item.content

        if current_item:
            result.append(current_item)

        new_text = self.text
        for res in result:
            new_text = new_text.replace(f"<_audio_>{res['audio']}</_audio_>", "")

        if new_text.strip(): 
            result.append({"text": new_text.strip()})

        for res in result:
            if "audio" in res and not res["audio"].startswith("data:audio/"):
                # Validate that the content looks like a file path
                audio_path = res["audio"].strip()
                if self._is_valid_file_path(audio_path):
                    try:
                        res["audio"] = Audio.load_audio_from_path(
                            audio_path, only_content=True
                        )
                    except Exception:
                        # If loading fails, keep the original path
                        pass

        return result

    def _is_valid_file_path(self, path: str) -> bool:
        """Check if the file path exists"""
        try:
            return os.path.exists(path.strip())
        except:
            return False



from typing import Optional

from .schema import LLMModel
from .registry import ModelRegistry


class PricingManager:
    """模型价格管理器"""
    
    def __init__(self, registry: ModelRegistry):
        self.registry = registry
    
    def update_input_price(self, name: str, price: float) -> bool:
        """
        更新模型输入价格
        
        Args:
            name: 模型名称
            price: 新的输入价格（美元/百万tokens），必须大于等于0
            
        Returns:
            bool: 是否成功更新
            
        Raises:
            ValueError: 如果价格为负数
        """
        if price < 0:
            raise ValueError("价格不能为负数")
        
        model = self.registry.get(name)
        if not model:
            return False
        
        model.input_price = float(price)
        self.registry.add_or_update(model)
        return True
    
    def update_output_price(self, name: str, price: float) -> bool:
        """
        更新模型输出价格
        
        Args:
            name: 模型名称
            price: 新的输出价格（美元/百万tokens），必须大于等于0
            
        Returns:
            bool: 是否成功更新
            
        Raises:
            ValueError: 如果价格为负数
        """
        if price < 0:
            raise ValueError("价格不能为负数")
        
        model = self.registry.get(name)
        if not model:
            return False
        
        model.output_price = float(price)
        self.registry.add_or_update(model)
        return True
    
    def update_prices(self, name: str, input_price: Optional[float] = None, 
                     output_price: Optional[float] = None) -> bool:
        """
        同时更新输入和输出价格
        
        Args:
            name: 模型名称
            input_price: 新的输入价格（可选）
            output_price: 新的输出价格（可选）
            
        Returns:
            bool: 是否成功更新
            
        Raises:
            ValueError: 如果任一价格为负数
        """
        if input_price is not None and input_price < 0:
            raise ValueError("输入价格不能为负数")
        if output_price is not None and output_price < 0:
            raise ValueError("输出价格不能为负数")
        
        model = self.registry.get(name)
        if not model:
            return False
        
        updated = False
        if input_price is not None:
            model.input_price = float(input_price)
            updated = True
        if output_price is not None:
            model.output_price = float(output_price)
            updated = True
        
        if updated:
            self.registry.add_or_update(model)
        
        return updated
    
    def get_cost_estimate(self, name: str, input_tokens: int, output_tokens: int) -> Optional[float]:
        """
        估算使用成本
        
        Args:
            name: 模型名称
            input_tokens: 输入 token 数量
            output_tokens: 输出 token 数量
            
        Returns:
            估算的总成本（美元），如果模型不存在则返回 None
        """
        model = self.registry.get(name)
        if not model:
            return None
        
        # 计算成本：价格单位是 美元/百万tokens
        input_cost = (input_tokens / 1_000_000) * model.input_price
        output_cost = (output_tokens / 1_000_000) * model.output_price
        
        return input_cost + output_cost 




"""
Unit tests for the MediaExtractor module
"""

import pytest
import tempfile
from pathlib import Path
from PIL import Image

from ..extractors import MediaExtractor, extract_image_paths, extract_audio_paths, convert_image_paths_from
from ..processors import ProcessorError


# Module-level fixtures for sharing between test classes
@pytest.fixture
def sample_images(tmp_path):
    """Create sample image files for testing"""
    images = {}
    
    # Create a red image
    img1_path = tmp_path / "red.png"
    img1 = Image.new('RGB', (10, 10), color='red')
    img1.save(img1_path)
    images['red'] = img1_path
    
    # Create a blue image
    img2_path = tmp_path / "blue.jpg"
    img2 = Image.new('RGB', (10, 10), color='blue')
    img2.save(img2_path)
    images['blue'] = img2_path
    
    return images


@pytest.fixture
def sample_audio(tmp_path):
    """Create sample audio files for testing"""
    audio_path = tmp_path / "test.mp3"
    audio_path.write_bytes(b"fake mp3 content" * 100)
    return audio_path


class TestMediaExtractor:
    """Test cases for MediaExtractor"""
    
    def test_extract_images_as_paths(self, sample_images):
        """Test extracting image paths without conversion"""
        extractor = MediaExtractor()
        
        text = f"<_image_>{sample_images['red']}</_image_> and <_image_>{sample_images['blue']}</_image_>"
        
        paths = extractor.extract_images(text, to_base64=False)
        assert len(paths) == 2
        assert str(sample_images['red']) in paths
        assert str(sample_images['blue']) in paths
    
    def test_extract_images_as_base64(self, sample_images):
        """Test extracting images and converting to base64"""
        extractor = MediaExtractor()
        
        text = f"<_image_>{sample_images['red']}</_image_>"
        
        results = extractor.extract_images(text, to_base64=True)
        assert len(results) == 1
        assert results[0].startswith("data:image/png;base64,")
    
    def test_extract_images_with_positions(self, sample_images):
        """Test extracting images with position information"""
        extractor = MediaExtractor()
        
        text = f"Start <_image_>{sample_images['red']}</_image_> middle <_image_>{sample_images['blue']}</_image_> end"
        
        results = extractor.extract_images(text, to_base64=False, return_positions=True)
        assert len(results) == 2
        
        # Check first image
        assert results[0]['content'] == str(sample_images['red'])
        assert results[0]['tag_name'] == '_image_'
        assert 'start' in results[0]
        assert 'end' in results[0]
        
        # Check positions make sense
        assert results[0]['start'] < results[0]['end']
        assert results[1]['start'] > results[0]['end']
    
    def test_extract_images_already_base64(self):
        """Test handling images that are already base64"""
        extractor = MediaExtractor()
        
        base64_data = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNkYPhfDwAChwGA60e6kgAAAABJRU5ErkJggg=="
        text = f"<_image_>{base64_data}</_image_>"
        
        results = extractor.extract_images(text, to_base64=True)
        assert len(results) == 1
        assert results[0] == base64_data  # Should return as-is
    
    def test_extract_images_custom_tags(self, sample_images):
        """Test extracting images with custom tag names"""
        extractor = MediaExtractor(image_tags=['img', 'photo'])
        
        text = f"<img>{sample_images['red']}</img> and <photo>{sample_images['blue']}</photo>"
        
        paths = extractor.extract_images(text, to_base64=False)
        assert len(paths) == 2
    
    def test_extract_images_error_handling(self, tmp_path):
        """Test error handling for invalid image paths"""
        # Test strict mode
        extractor_strict = MediaExtractor(strict_mode=True)
        text = "<_image_>/path/does/not/exist.jpg</_image_>"
        
        with pytest.raises(ProcessorError):
            extractor_strict.extract_images(text, to_base64=True)
        
        # Test non-strict mode (should log error but not raise)
        extractor_lenient = MediaExtractor(strict_mode=False)
        results = extractor_lenient.extract_images(text, to_base64=True)
        assert len(results) == 0  # Invalid path is skipped
    
    def test_extract_audio(self, sample_audio):
        """Test extracting audio files"""
        extractor = MediaExtractor()
        
        text = f"<_audio_>{sample_audio}</_audio_>"
        
        # Test path extraction
        paths = extractor.extract_audio(text, to_base64=False)
        assert len(paths) == 1
        assert paths[0] == str(sample_audio)
        
        # Test base64 conversion
        results = extractor.extract_audio(text, to_base64=True)
        assert len(results) == 1
        assert results[0].startswith("data:audio/mpeg;base64,")
    
    def test_extract_audio_with_positions(self, sample_audio):
        """Test extracting audio with position information"""
        extractor = MediaExtractor()
        
        text = f"Audio here: <_audio_>{sample_audio}</_audio_> done"
        
        results = extractor.extract_audio(text, to_base64=False, return_positions=True)
        assert len(results) == 1
        assert results[0]['content'] == str(sample_audio)
        assert results[0]['tag_name'] == '_audio_'
        assert 'start' in results[0]
        assert 'end' in results[0]
    
    def test_convert_tag_format(self, sample_images):
        """Test converting between tag formats"""
        extractor = MediaExtractor()
        
        # Convert from <img> to <_image_>
        text = f"<img>{sample_images['red']}</img> text <img>{sample_images['blue']}</img>"
        converted = extractor.convert_tag_format(text, "<img>", "<_image_>", media_type='image')
        
        assert "<img>" not in converted
        assert "<_image_>" in converted
        assert str(sample_images['red']) in converted
        assert str(sample_images['blue']) in converted
    
    def test_to_content_list(self, sample_images, sample_audio):
        """Test converting text to content list"""
        extractor = MediaExtractor()
        
        text = f"""
        This is some text.
        <_image_>{sample_images['red']}</_image_>
        More text here.
        <_audio_>{sample_audio}</_audio_>
        Final text.
        """
        
        content_list = extractor.to_content_list(text)
        
        # Should have alternating text and media
        assert len(content_list) == 5
        assert 'text' in content_list[0]
        assert 'image' in content_list[1]
        assert 'text' in content_list[2]
        assert 'audio' in content_list[3]
        assert 'text' in content_list[4]
        
        # Check that image was converted to base64
        assert content_list[1]['image'].startswith('data:image/')
        assert content_list[3]['audio'].startswith('data:audio/')
    
    def test_to_content_list_no_text_between(self, sample_images):
        """Test content list when media tags are adjacent"""
        extractor = MediaExtractor()
        
        text = f"<_image_>{sample_images['red']}</_image_><_image_>{sample_images['blue']}</_image_>"
        
        content_list = extractor.to_content_list(text)
        
        # Should only have images, no text
        assert len(content_list) == 2
        assert 'image' in content_list[0]
        assert 'image' in content_list[1]
    
    def test_remove_media_tags(self, sample_images, sample_audio):
        """Test removing all media tags from text"""
        extractor = MediaExtractor()
        
        text = f"""
        Text before <_image_>{sample_images['red']}</_image_> text middle 
        <_audio_>{sample_audio}</_audio_> text after
        """
        
        cleaned = extractor.remove_media_tags(text)
        
        assert "<_image_>" not in cleaned
        assert "<_audio_>" not in cleaned
        assert str(sample_images['red']) not in cleaned
        assert str(sample_audio) not in cleaned
        assert "Text before" in cleaned
        assert "text middle" in cleaned
        assert "text after" in cleaned
    
    def test_validate_media_paths(self, sample_images, tmp_path):
        """Test validating media file paths"""
        extractor = MediaExtractor()
        
        # Mix of valid and invalid paths
        text = f"""
        <_image_>{sample_images['red']}</_image_>
        <_image_>/invalid/path.jpg</_image_>
        <_image_>data:image/png;base64,abc123</_image_>
        <_audio_>/another/invalid.mp3</_audio_>
        """
        
        validation = extractor.validate_media_paths(text)
        
        assert str(sample_images['red']) in validation['valid']
        assert '/invalid/path.jpg' in validation['invalid']
        assert '/another/invalid.mp3' in validation['invalid']
        # Base64 data should be skipped, not validated
        assert 'data:image/png;base64,abc123' not in validation['valid']
        assert 'data:image/png;base64,abc123' not in validation['invalid']
    
    def test_max_file_size_configuration(self, tmp_path):
        """Test configuring maximum file size"""
        # Create a large file
        large_file = tmp_path / "large.png"
        large_file.write_bytes(b"x" * 1000)
        
        # Extractor with small file size limit
        extractor = MediaExtractor(max_file_size=500)
        
        text = f"<_image_>{large_file}</_image_>"
        
        # Should fail in strict mode
        extractor_strict = MediaExtractor(strict_mode=True, max_file_size=500)
        with pytest.raises(ProcessorError):
            extractor_strict.extract_images(text, to_base64=True)


class TestBackwardCompatibility:
    """Test backward compatibility functions"""
    
    def test_extract_image_paths(self, sample_images):
        """Test backward compatible extract_image_paths function"""
        text = f"<_image_>{sample_images['red']}</_image_>"
        
        # Test without base64 conversion
        paths = extract_image_paths(text, to_base64=False)
        assert len(paths) == 1
        assert paths[0] == str(sample_images['red'])
        
        # Test with base64 conversion
        results = extract_image_paths(text, to_base64=True)
        assert len(results) == 1
        assert results[0].startswith("data:image/")
    
    def test_extract_audio_paths(self, sample_audio):
        """Test backward compatible extract_audio_paths function"""
        text = f"<_audio_>{sample_audio}</_audio_>"
        
        paths = extract_audio_paths(text, to_base64=False)
        assert len(paths) == 1
        assert paths[0] == str(sample_audio)
    
    def test_convert_image_paths_from(self, sample_images):
        """Test backward compatible convert_image_paths_from function"""
        text = f"<img>{sample_images['red']}</img>"
        
        converted = convert_image_paths_from(text, "<img>", "</img>")
        assert "<_image_>" in converted
        assert str(sample_images['red']) in converted





from typing import List, Dict, Any
import byzerllm
from loguru import logger
from pydantic import BaseModel
from byzerllm.apps.utils import TagExtractor, Tag
import importlib
import json
import hashlib
import os
from pathlib import Path


def read_alpaca_zh():
    with importlib.resources.path(
        "byzerllm.apps.byzer_storage", "alpaca_zh.json"
    ) as json_path:
        with open(json_path, "r", encoding="utf-8") as f:
            return json.load(f)


class QAPair(BaseModel):
    question: str
    answer: str


@byzerllm.prompt()
def _convert_pretrain_text_to_instruction(text: str, num: int = 10) -> str:
    """
    根据提供的信息，生成多个相关的问题，这些问题的回答，最终要能覆盖里面所有的信息,生成 {{num}} 个。

    下面是一些信息：

    {{text}}

    现在请开始生成问题，请将每个问题使用<_question_></_question_>标签包裹，每个回答使用<_answer_></_answer_>标签包裹，最后
    每个一组问题和回答使用<_group_></_group_>标签包裹。
    """


@byzerllm.prompt()
def get_more(text: str, questions_and_answers: str, num: int = 10) -> str:
    """
    下面是一些问答信息：

    {{text}}

    下面是已有的问题和回答：

    {{questions_and_answers}}

    下面是已经生成的问题和回答，请继续生成{{num}}个问题和回答，不要和前面的重复，请将每个问题使用<_question_></_question_>标签包裹，每个回答使用<_answer_></_answer_>标签包裹，最后
    每个一组问题和回答使用<_group_></_group_>标签包裹。
    """


def generate_cache_key(text: str) -> str:
    return hashlib.md5(text.encode()).hexdigest()


def get_cache_path() -> Path:
    home = Path.home()
    cache_dir = home / ".byzerllm" / "cache" / "llm_generated_data"
    cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir


def save_to_cache(key: str, data: List[QAPair]):
    cache_path = get_cache_path() / f"{key}.json"
    with cache_path.open("w", encoding="utf-8") as f:
        json.dump([qa.dict() for qa in data], f, ensure_ascii=False, indent=2)


def load_from_cache(key: str) -> List[QAPair]:
    cache_path = get_cache_path() / f"{key}.json"
    if cache_path.exists():
        with cache_path.open("r", encoding="utf-8") as f:
            data = json.load(f)
        return [QAPair(**item) for item in data]
    return None


def to_qa_pairs(text: str, llm, num: int = 20) -> List[QAPair]:
    cache_key = generate_cache_key(text)
    cached_result = load_from_cache(cache_key)

    if cached_result:
        logger.info(f"Using cached result for text: {text[:100]}...")
        return cached_result

    logger.info(f"Starting to generate QA pairs for text: {text[:100]}...")

    logger.info("Generating initial QA pairs...")
    v = _convert_pretrain_text_to_instruction.with_llm(llm).run(text)
    logger.info(f"Initial QA pairs generated. Length: {len(v)}")

    max_turns = int(num / 10)
    if max_turns < 1:
        max_turns = 1
    logger.info(f"Will generate additional QA pairs for {max_turns} turns")

    turn = 0
    while turn < max_turns:
        turn += 1
        logger.info(f"Generating additional QA pairs (turn {turn}/{max_turns})...")
        t = get_more.with_llm(llm).run(text, v)
        if t:
            v += t
            logger.info(f"Additional QA pairs generated. New total length: {len(v)}")

    logger.info(f"Extracting QA pairs from generated text...{v}")
    root_tag = TagExtractor(v).extract()
    qa_pairs = []
    for i, item in enumerate(root_tag.content):
        qas = item.content
        if len(qas) == 2:
            if (
                qas[0].start_tag == "<_question_>"
                and qas[0].end_tag == "</_question_>"
                and qas[1].start_tag == "<_answer_>"
                and qas[1].end_tag == "</_answer_>"
            ):
                qa_pairs.append(QAPair(question=qas[0].content, answer=qas[1].content))
                logger.debug(
                    f"Extracted QA pair {i+1}: Q: {qas[0].content[:50]}... A: {qas[1].content[:50]}..."
                )
            else:
                logger.warning(
                    f"Skipping item {i+1} due to incorrect tags: {qas[0].start_tag}, {qas[0].end_tag}, {qas[1].start_tag}, {qas[1].end_tag}"
                )
        else:
            logger.warning(
                f"Skipping item {i+1} due to incorrect number of elements: {len(qas)}"
            )

    logger.info(f"Extracted {len(qa_pairs)} valid QA pairs.")

    # Save the result to cache
    save_to_cache(cache_key, qa_pairs)

    return qa_pairs



from prometheus_client import CollectorRegistry, Gauge,Counter, pushadd_to_gateway
from byzerllm.utils.config import get_mlsql_config_pushgateway_address,get_mlsql_config
from typing import Union,Dict
import ray

class Metric:

    def __init__(self):
        self.registry = CollectorRegistry()
        config = get_mlsql_config()
        self.metric_enabled = False
        self.pushgateway_address = None
        if config is not None:
            self.pushgateway_address = ray.get(config.getitem.remote("spark.mlsql.pushgateway.address",None))
            self.metric_enabled = True

        self.gauges = {}
        self.counters = {}        

    def inc(self, name:str,value: Union[int, float] = 1.0, tags: Dict[str, str] = None):
        if not self.metric_enabled:
            return
        if name not in self.counters:
            self.counters[name] = Counter(name, '', registry=self.registry)                    
        self.counters[name].inc(value)   

    def push(self):
        if not self.metric_enabled:
            return
        if self.pushgateway_address is not None:
            pushadd_to_gateway(self.pushgateway_address, job='pushgateway', registry=self.registry)      
        
    




"""
Core Config Module

Provides centralized configuration and memory management for auto-coder.

This module has been refactored into multiple specialized modules for better
maintainability and separation of concerns:

- models.py: Core data structures
- base_manager.py: Base persistence functionality  
- config_manager.py: Configuration management
- file_manager.py: File and file group management
- exclude_manager.py: Exclude patterns management
- lib_manager.py: Library management
- conversation_manager.py: Conversation management
- main_manager.py: Combined MemoryManager class
- compatibility.py: Backward compatibility functions
"""

# Import core classes and functions
from .models import CoreMemory
from .main_manager import MemoryManager, get_memory_manager
from .compatibility import (
    save_memory,
    save_memory_with_new_memory,
    load_memory,
    get_memory,
)

# Export all public interfaces
__all__ = [
    'MemoryManager',
    'CoreMemory', 
    'get_memory_manager',
    'save_memory',
    'save_memory_with_new_memory',
    'load_memory',
    'get_memory',
]



"""
File and file group management functionality.

This module provides methods for managing current files, file groups,
group information, and active groups.
"""

from typing import Dict, Any, List
from .models import FileList, GroupsDict, GroupsInfoDict


class FileManagerMixin:
    """Mixin class providing file and file group management functionality."""
    
    # ===== File Management =====
    def get_current_files(self) -> FileList:
        """Get current files list."""
        return self._memory.current_files.get("files", [])
    
    def set_current_files(self, files: FileList):
        """Set current files list."""
        self._memory.current_files["files"] = files
        self.save_memory()
    
    def add_files(self, files: FileList) -> FileList:
        """Add files to current files list (avoiding duplicates)."""
        current_files = set(self._memory.current_files.get("files", []))
        new_files = [f for f in files if f not in current_files]
        if new_files:
            self._memory.current_files["files"].extend(new_files)
            self.save_memory()
        return new_files
    
    def remove_files(self, files: FileList) -> FileList:
        """Remove files from current files list."""
        current_files = self._memory.current_files.get("files", [])
        removed_files = []
        for file in files:
            if file in current_files:
                current_files.remove(file)
                removed_files.append(file)
        if removed_files:
            self.save_memory()
        return removed_files
    
    def clear_files(self):
        """Clear all current files."""
        self._memory.current_files["files"] = []
        self.save_memory()
    
    # ===== File Groups Management =====
    def get_file_groups(self) -> GroupsDict:
        """Get all file groups."""
        return self._memory.current_files.get("groups", {})
    
    def get_file_group(self, name: str) -> FileList:
        """Get files in a specific group."""
        return self._memory.current_files.get("groups", {}).get(name, [])
    
    def set_file_group(self, name: str, files: FileList):
        """Set files for a group."""
        if "groups" not in self._memory.current_files:
            self._memory.current_files["groups"] = {}
        self._memory.current_files["groups"][name] = files
        self.save_memory()
    
    def add_file_group(self, name: str, files: FileList):
        """Add a new file group with current files or specified files."""
        if not files:
            files = self.get_current_files().copy()
        self.set_file_group(name, files)
    
    def delete_file_group(self, name: str) -> bool:
        """Delete a file group."""
        groups = self._memory.current_files.get("groups", {})
        if name in groups:
            del groups[name]
            # Also remove from groups_info and current_groups
            groups_info = self._memory.current_files.get("groups_info", {})
            if name in groups_info:
                del groups_info[name]
            current_groups = self._memory.current_files.get("current_groups", [])
            if name in current_groups:
                current_groups.remove(name)
            self.save_memory()
            return True
        return False
    
    def has_file_group(self, name: str) -> bool:
        """Check if file group exists."""
        return name in self._memory.current_files.get("groups", {})
    
    # ===== File Groups Info Management =====
    def get_groups_info(self) -> GroupsInfoDict:
        """Get all groups info."""
        return self._memory.current_files.get("groups_info", {})
    
    def get_group_info(self, name: str) -> Dict[str, Any]:
        """Get info for a specific group."""
        return self._memory.current_files.get("groups_info", {}).get(name, {})
    
    def set_group_info(self, name: str, info: Dict[str, Any]):
        """Set info for a group."""
        if "groups_info" not in self._memory.current_files:
            self._memory.current_files["groups_info"] = {}
        self._memory.current_files["groups_info"][name] = info
        self.save_memory()
    
    def set_group_query_prefix(self, name: str, query_prefix: str):
        """Set query prefix for a group."""
        info = self.get_group_info(name)
        info["query_prefix"] = query_prefix
        self.set_group_info(name, info)
    
    def get_group_query_prefix(self, name: str) -> str:
        """Get query prefix for a group."""
        return self.get_group_info(name).get("query_prefix", "")
    
    # ===== Current Active Groups Management =====
    def get_current_groups(self) -> List[str]:
        """Get currently active groups."""
        return self._memory.current_files.get("current_groups", [])
    
    def set_current_groups(self, groups: List[str]):
        """Set currently active groups."""
        self._memory.current_files["current_groups"] = groups
        self.save_memory()
    
    def add_current_group(self, group_name: str):
        """Add group to current active groups."""
        current_groups = self._memory.current_files.get("current_groups", [])
        if group_name not in current_groups:
            current_groups.append(group_name)
            self.save_memory()
    
    def remove_current_group(self, group_name: str) -> bool:
        """Remove group from current active groups."""
        current_groups = self._memory.current_files.get("current_groups", [])
        if group_name in current_groups:
            current_groups.remove(group_name)
            self.save_memory()
            return True
        return False
    
    def clear_current_groups(self):
        """Clear all current active groups."""
        self._memory.current_files["current_groups"] = []
        self.save_memory()
    
    def merge_groups_to_current_files(self, group_names: List[str]) -> List[str]:
        """Merge files from multiple groups and set as current files."""
        merged_files = set()
        existing_groups = []
        missing_groups = []
        
        groups = self.get_file_groups()
        for group_name in group_names:
            if group_name in groups:
                merged_files.update(groups[group_name])
                existing_groups.append(group_name)
            else:
                missing_groups.append(group_name)
        
        if merged_files:
            self.set_current_files(list(merged_files))
            self.set_current_groups(existing_groups)
        
        return missing_groups
    
    # ===== Legacy File Group Helpers =====
    def get_groups(self) -> GroupsDict:
        """Get file groups (legacy method for backward compatibility)."""
        return self.get_file_groups()
    
    def set_group(self, name: str, files: FileList):
        """Set a file group (legacy method for backward compatibility)."""
        self.set_file_group(name, files)



import os
import time
import uvicorn
from fastapi import Request
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, StreamingResponse, Response

from byzerllm.log import init_logger
from byzerllm.utils import random_uuid
from byzerllm.version import __version__ as version
from byzerllm.utils.client import ByzerLLM, LLMRequest
from byzerllm.utils.client.entrypoints.openai.serving_chat import OpenAIServingChat
from byzerllm.utils.client.entrypoints.openai.serving_completion import OpenAIServingCompletion
from byzerllm.utils.client.entrypoints.openai.protocol import (
    ModelList,
    ModelCard,
    ModelPermission,
    ChatCompletionRequest,
    ErrorResponse,
    CompletionRequest,
    EmbeddingCompletionRequest,
    EmbeddingResponseData,
    EmbeddingResponse,
    UsageInfo
)
from pydantic import BaseModel
from typing import List

logger = init_logger(__name__)

llm_client: ByzerLLM = None
openai_serving_chat: OpenAIServingChat = None
openai_serving_completion: OpenAIServingCompletion = None

TIMEOUT_KEEP_ALIVE = 5  # seconds
# timeout in 10 minutes. Streaming can take longer than 3 min
TIMEOUT = float(os.environ.get("BYZERLLM_APISERVER_HTTP_TIMEOUT", 600))

router_app = FastAPI()


@router_app.get("/health")
async def health() -> Response:
    """Health check."""
    return Response(status_code=200)


@router_app.get("/v1/models")
async def show_available_models():
    models = await openai_serving_chat.show_available_models()
    return JSONResponse(content=models.model_dump())


@router_app.get("/version")
async def show_version():
    return JSONResponse(content={"version": version})


@router_app.get("/v1/models", response_model=ModelList)
async def models() -> ModelList:
    """Show available models. Right now we only have one model."""
    model_cards = [
        ModelCard(
            id="",
            root="",
            permission=[ModelPermission()]
        )
    ]
    return ModelList(data=model_cards)


@router_app.post("/v1/completions")
async def create_completion(
        body: CompletionRequest,
        request: Request
):
    generator = await openai_serving_completion.create_completion(body, request)
    if isinstance(generator, ErrorResponse):
        return JSONResponse(
            content=generator.model_dump(),
            status_code=generator.code
        )
    if body.stream:
        return StreamingResponse(
            content=generator,
            media_type="text/event-stream"
        )
    else:
        return JSONResponse(content=generator.model_dump())


@router_app.post("/v1/chat/completions")
async def create_chat_completion(
        body: ChatCompletionRequest,
        request: Request,
):
    """Completion API similar to OpenAI's API.

    See  https://platform.openai.com/docs/api-reference/chat/create
    for the API specification. This API mimics the OpenAI ChatCompletion API.

    NOTE: Currently we do not support the following features:
        - function_call (Users should implement this by themselves)
        - logit_bias (to be supported by vLLM engine)
    """
    # async with async_timeout.timeout(TIMEOUT):

    generator = await openai_serving_chat.create_chat_completion(body, request)
    if isinstance(generator, ErrorResponse):
        return JSONResponse(
            content=generator.model_dump(),
            status_code=generator.code
        )
    if body.stream:
        return StreamingResponse(
            content=generator,
            media_type="text/event-stream"
        )
    else:
        return JSONResponse(content=generator.model_dump())


@router_app.post("/v1/embeddings")
async def embed(body: EmbeddingCompletionRequest):
    """Generate embeddings for given input text.
    
    Args:
        body: The embedding request containing input text and parameters.
        
    Returns:
        EmbeddingResponse with embeddings and usage statistics.
    """
    embedding_id = f"embed-{random_uuid()}"
    
    # Handle both string and list inputs
    inputs = body.input if isinstance(body.input, list) else [body.input]
    
    # Generate embeddings for each input
    results_list = []
    for text in inputs:
        result = llm_client.emb(body.model, request=LLMRequest(instruction=text))
        results_list.extend(result)

    # Build response data
    data = [
        EmbeddingResponseData(
            embedding=result.output,
            index=i,
            object="embedding"
        )
        for i, result in enumerate(results_list)
    ]
    
    # Calculate token usage (simplified)
    token_count = sum(len(str(input).split()) for input in inputs)
    
    return EmbeddingResponse(
        data=data,
        model=body.model,
        object="list",
        usage=UsageInfo(
            prompt_tokens=token_count,
            total_tokens=token_count
        ),
        created=int(time.time()),
        id=embedding_id
    )

class ServerArgs(BaseModel):
    host: str = None
    port: int = 8000
    uvicorn_log_level: str = "info"
    allow_credentials: bool = False
    allowed_origins: List[str] = ["*"]  
    allowed_methods: List[str] = ["*"]
    allowed_headers: List[str] = ["*"]
    api_key: str = None
    served_model_name: str = None
    prompt_template: str = None
    response_role: str = "assistant"
    ssl_keyfile: str = None
    ssl_certfile: str = None

def serve(llm:ByzerLLM, args: ServerArgs):
    
    logger.info(f"ByzerLLM API server version {version}")
    logger.info(f"args: {args}")

    router_app.add_middleware(
        CORSMiddleware,
        allow_origins=args.allowed_origins,
        allow_credentials=args.allow_credentials,
        allow_methods=args.allowed_methods,
        allow_headers=args.allowed_headers,
    )
    
    if token := os.environ.get("BYZERLLM_API_KEY") or args.api_key:

        @router_app.middleware("http")
        async def authentication(request: Request, call_next):
            if not request.url.path.startswith("/v1"):
                return await call_next(request)
            if request.headers.get("Authorization") != "Bearer " + token:
                return JSONResponse(
                    content={"error": "Unauthorized"},
                    status_code=401
                )
            return await call_next(request)

    # Register labels for metrics
    # add_global_metrics_labels(model_name=engine_args.model)
    global llm_client
    llm_client = llm
    
    global openai_serving_chat
    openai_serving_chat = OpenAIServingChat(
        llm_client=llm_client,
        response_role=args.response_role,
        server_model_name=args.served_model_name,
        prompt_template=args.prompt_template
    )
    global openai_serving_completion
    openai_serving_completion = OpenAIServingCompletion(
        llm_client=llm_client,
        server_model_name=args.served_model_name,
        prompt_template=args.prompt_template
    )

    uvicorn.run(
        router_app,
        host=args.host,
        port=args.port,
        log_level=args.uvicorn_log_level,
        timeout_keep_alive=TIMEOUT_KEEP_ALIVE,
        ssl_keyfile=args.ssl_keyfile,
        ssl_certfile=args.ssl_certfile
    )



"""
Exclude management functionality.

This module provides methods for managing excluded files and directories.
"""

from typing import List
from .models import FileList


class ExcludeManagerMixin:
    """Mixin class providing exclude management functionality."""
    
    # ===== Exclude Files Management =====
    def get_exclude_files(self) -> FileList:
        """Get exclude file patterns."""
        return self._memory.exclude_files.copy()
    
    def add_exclude_files(self, patterns: FileList) -> FileList:
        """Add exclude file patterns (avoiding duplicates)."""
        existing_patterns = set(self._memory.exclude_files)
        new_patterns = [p for p in patterns if p not in existing_patterns]
        if new_patterns:
            self._memory.exclude_files.extend(new_patterns)
            self.save_memory()
        return new_patterns
    
    def remove_exclude_files(self, patterns: FileList) -> FileList:
        """Remove exclude file patterns."""
        removed_patterns = []
        for pattern in patterns:
            if pattern in self._memory.exclude_files:
                self._memory.exclude_files.remove(pattern)
                removed_patterns.append(pattern)
        if removed_patterns:
            self.save_memory()
        return removed_patterns
    
    def clear_exclude_files(self):
        """Clear all exclude file patterns."""
        self._memory.exclude_files = []
        self.save_memory()
    
    # ===== Exclude Directories Management =====
    def get_exclude_dirs(self) -> List[str]:
        """Get exclude directories."""
        return self._memory.exclude_dirs.copy()
    
    def add_exclude_dirs(self, dirs: List[str]) -> List[str]:
        """Add exclude directories (avoiding duplicates)."""
        existing_dirs = set(self._memory.exclude_dirs)
        new_dirs = [d for d in dirs if d not in existing_dirs]
        if new_dirs:
            self._memory.exclude_dirs.extend(new_dirs)
            self.save_memory()
        return new_dirs
    
    def remove_exclude_dirs(self, dirs: List[str]) -> List[str]:
        """Remove exclude directories."""
        removed_dirs = []
        for dir_name in dirs:
            if dir_name in self._memory.exclude_dirs:
                self._memory.exclude_dirs.remove(dir_name)
                removed_dirs.append(dir_name)
        if removed_dirs:
            self.save_memory()
        return removed_dirs
    
    def clear_exclude_dirs(self):
        """Clear all exclude directories."""
        self._memory.exclude_dirs = []
        self.save_memory()



import asyncio
import ray
import time
import os
import tarfile
import tempfile
import uuid
from .utils import print_flush

async def producer(items,udf_name, queue):    
    for item in items:        
        await queue.put(ray.get(item))
    await queue.put(None)

async def worker(queue,udf_name,tf_path,total_count):
    count = 0
    with open(tf_path, "wb") as tf:
        while True:
            item = await queue.get()
            if item is None:
                # Signal to exit when queue is empty
                break
            tf.write(item["value"]) 
            if count % 1000 == 0:
                print_flush(f"MODEL[{udf_name}] UDFWorker pull model: {float(count)/total_count*100}%")
            count += 1  

async def _transfer_from_ob(udf_name, model_refs,target_dir):
    queue = asyncio.Queue(1000)
    
    tf_path = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))      

    worker_task = asyncio.create_task(worker(queue,udf_name,tf_path,len(model_refs)))
    producer_task = asyncio.create_task(producer(model_refs, udf_name,queue))    
    
    await asyncio.gather(producer_task,worker_task)
    
    with open(tf_path, "rb") as tf:
        tt = tarfile.open(tf.name, mode="r:")
        tt.extractall(target_dir)
        tt.close()
    os.remove(tf_path)

def block_transfer_from_ob(udf_name, model_refs,target_dir):
    tf_path = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))  
    count = 0
    total_count = len(model_refs)
    with open(tf_path, "wb") as tf:
        for item_ref in model_refs:
            item = ray.get(item_ref)            
            tf.write(item["value"]) 
            if count % 1000 == 0:
                print_flush(f"MODEL[{udf_name}] UDFWorker pull model: {float(count)/total_count*100}%")
            count += 1  
    with open(tf_path, "rb") as tf:
        tt = tarfile.open(tf.name, mode="r:")
        tt.extractall(target_dir)
        tt.close()
    os.remove(tf_path)        


def transfer_from_ob(udf_name,model_refs,model_dir):
    print_flush(f"[{udf_name}] model_refs:{len(model_refs)} model_dir:{model_dir}")
    time1 = time.time()
    # loop = asyncio.get_event_loop()
    # loop.run_until_complete(_transfer_from_ob(udf_name,model_refs,model_dir))            
    block_transfer_from_ob(udf_name,model_refs,model_dir)
    print_flush(f"[{udf_name}] UDFWorker pull model from object store cost {time.time() - time1} seconds")



"""
Main MemoryManager class combining all functionality.

This module provides the main MemoryManager class that combines all
functionality from the various manager mixins.
"""

from typing import Optional
from .base_manager import BaseMemoryManager
from .config_manager import ConfigManagerMixin
from .file_manager import FileManagerMixin
from .exclude_manager import ExcludeManagerMixin
from .lib_manager import LibManagerMixin
from .conversation_manager import ConversationManagerMixin


class MemoryManager(
    BaseMemoryManager,
    ConfigManagerMixin,
    FileManagerMixin,
    ExcludeManagerMixin,
    LibManagerMixin,
    ConversationManagerMixin
):
    """
    Complete memory manager for auto-coder sessions.
    
    This class combines all functionality from various manager mixins:
    - BaseMemoryManager: Core persistence and singleton functionality
    - ConfigManagerMixin: Configuration management
    - FileManagerMixin: File and file group management
    - ExcludeManagerMixin: Exclude patterns management
    - LibManagerMixin: Library management
    - ConversationManagerMixin: Conversation history management
    
    Provides thread-safe persistence of configuration and session data.
    """
    
    def __init__(self, project_root: Optional[str] = None):
        """
        Initialize memory manager with all functionality.
        
        Args:
            project_root: Project root directory. If None, uses current working directory.
        """
        super().__init__(project_root)


def get_memory_manager(project_root: Optional[str] = None) -> MemoryManager:
    """
    Get memory manager instance.
    
    Args:
        project_root: Project root directory. If None, uses current working directory.
        
    Returns:
        MemoryManager instance
    """
    return MemoryManager.get_instance(project_root)



import os

import torch

from byzerllm.stable_diffusion.config import stableDiffusionConfig


def hf_diffusers_cache_dir():
    cache_dir = os.path.join(stableDiffusionConfig.get_temp_dir(), "diffusers")
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir


def hf_transformers_cache_dir():
    cache_dir = os.path.join(stableDiffusionConfig.get_temp_dir(), "transformers")
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir


def get_device():
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")
